from pydantic import Field, JsonValue

from cognite_toolkit._cdf_tk.resource_classes.base import ToolkitResource


class RobotCapabilityYAML(ToolkitResource):
    name: str = Field(description="Robot capability name.")
    external_id: str = Field(description="RobotCapability external id.", max_length=256)
    method: str = Field(description="RobotCapability method, used to call the right functionality on the robot.")
    input_schema: dict[str, JsonValue] | None = Field(
        default=None, description="Schema that defines what inputs are needed for configuring the action."
    )
    data_handling_schema: dict[str, JsonValue] | None = Field(
        default=None, description="Schema that defines how the data from a RobotCapability should be handled."
    )
    description: str | None = Field(
        default=None,
        description="Textual description of the RobotCapability.",
    )



from uuid import uuid4

from cognite.client.data_classes.capabilities import (
    Capability,
    DataModelInstancesAcl,
    DataModelsAcl,
    SpaceIDScope,
)
from cognite.client.exceptions import CogniteException

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.canvas import (
    CANVAS_INSTANCE_SPACE,
    Canvas,
    ContainerReferenceApply,
    FdmInstanceContainerReferenceApply,
)
from cognite_toolkit._cdf_tk.client.data_classes.migration import InstanceSource
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.commands._migrate.data_model import (
    INSTANCE_SOURCE_VIEW_ID,
    MODEL_ID,
    RESOURCE_VIEW_MAPPING_VIEW_ID,
)
from cognite_toolkit._cdf_tk.exceptions import AuthenticationError, ToolkitMigrationError
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning, LowSeverityWarning, MediumSeverityWarning
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.interactive_select import InteractiveCanvasSelect


class MigrationCanvasCommand(ToolkitCommand):
    canvas_schema_space = Canvas.get_source().space
    # Note sequences are not supported in Canvas, so we do not include them here.
    asset_centric_resource_types = frozenset({"asset", "event", "file", "timeseries"})

    def migrate_canvas(
        self,
        client: ToolkitClient,
        external_ids: list[str] | None = None,
        dry_run: bool = False,
        verbose: bool = False,
    ) -> None:
        self.validate_access(
            client,
            instance_spaces=[CANVAS_INSTANCE_SPACE],
            schema_spaces=[self.canvas_schema_space, INSTANCE_SOURCE_VIEW_ID.space],
        )
        self.validate_migration_model_available(client)
        external_ids = external_ids or InteractiveCanvasSelect(client).select_external_ids()
        if external_ids is None or not external_ids:
            self.console("No canvases selected for migration.")
            return
        action = "Would migrate" if dry_run else "Migrating"
        self.console(f"{action} {len(external_ids)} canvases.")
        for external_id in external_ids:
            self._migrate_single_canvas(client, external_id, dry_run=dry_run, verbose=verbose)

    def _migrate_single_canvas(
        self,
        client: ToolkitClient,
        external_id: str,
        dry_run: bool = False,
        verbose: bool = False,
    ) -> None:
        canvas = client.canvas.industrial.retrieve(external_id=external_id)
        if canvas is None:
            self.warn(MediumSeverityWarning(f"Canvas with external ID '{external_id}' not found. Skipping.. "))
            return
        update = canvas.as_write()
        to_migrate = [
            ref
            for ref in update.container_references
            if ref.container_reference_type in self.asset_centric_resource_types
        ]
        if not to_migrate:
            self.warn(
                LowSeverityWarning(
                    f"Canvas with name '{canvas.canvas.name}' does not have any asset-centric references. Skipping.. "
                )
            )
        if verbose:
            self.console(f"Found canvas: {canvas.canvas.name}")
        reference_ids = [ref.as_asset_centric_id() for ref in to_migrate]
        instance_sources = client.migration.instance_source.retrieve(reference_ids)
        source_by_reference_id = {source.as_asset_centric_id(): source for source in instance_sources}
        missing = set(reference_ids) - set(source_by_reference_id.keys())
        if missing:
            self.warn(
                HighSeverityWarning(
                    f"Canvas '{canvas.canvas.name}' has references to resources that are not been migrated: {humanize_collection(missing)}. Skipping.. "
                )
            )
            return
        if dry_run:
            self.console(
                f"Canvas '{canvas.canvas.name}' is ready for migration all {len(instance_sources)} references asset-centric resources found."
            )
            return
        if verbose:
            self.console(
                f"Migrating canvas '{canvas.canvas.name}' with {len(instance_sources)} references to asset-centric resources."
            )
        backup = canvas.as_write().create_backup()

        update.container_references = [
            ref
            for ref in update.container_references
            if ref.container_reference_type not in self.asset_centric_resource_types
        ]
        for ref in to_migrate:
            source = source_by_reference_id[ref.as_asset_centric_id()]
            fdm_ref = self.migrate_container_reference(ref, source, canvas.canvas.external_id)
            update.fdm_instance_container_references.append(fdm_ref)

        try:
            client.canvas.industrial.create(backup)
        except CogniteException as e:
            raise ToolkitMigrationError(f"Failed to create backup for canvas '{canvas.canvas.name}': {e!s}. ") from e
        try:
            client.canvas.industrial.update(update)
        except CogniteException as e:
            raise ToolkitMigrationError(
                f"Failed to migrate canvas '{canvas.canvas.name}': {e!s}. A backup was created with external ID '{backup.canvas.external_id}'."
            ) from e
        else:
            self.console(
                f'Canvas "{canvas.canvas.name}" migrated successfully with {len(to_migrate)} references to data model instances.'
            )

    @classmethod
    def migrate_container_reference(
        cls, reference: ContainerReferenceApply, source: InstanceSource, canvas_external_id: str
    ) -> FdmInstanceContainerReferenceApply:
        """Migrate a single container reference by replacing the asset-centric ID with the data model instance ID."""
        consumer_view = source.consumer_view()
        new_id = str(uuid4())
        new_external_id = f"{canvas_external_id}_{new_id}"
        return FdmInstanceContainerReferenceApply(
            external_id=new_external_id,
            id_=new_id,
            container_reference_type="fdmInstance",
            instance_space=source.space,
            instance_external_id=source.external_id,
            view_space=consumer_view.space,
            view_external_id=consumer_view.external_id,
            view_version=consumer_view.version,
            label=reference.label,
            properties_=reference.properties_,
            x=reference.x,
            y=reference.y,
            width=reference.width,
            height=reference.height,
            max_width=reference.max_width,
            max_height=reference.max_height,
        )

    @staticmethod
    def validate_access(
        client: ToolkitClient,
        instance_spaces: list[str] | None = None,
        schema_spaces: list[str] | None = None,
    ) -> None:
        required_capabilities: list[Capability] = []
        if instance_spaces is not None:
            required_capabilities.append(
                DataModelInstancesAcl(
                    actions=[
                        DataModelInstancesAcl.Action.Read,
                        DataModelInstancesAcl.Action.Write,
                        DataModelInstancesAcl.Action.Write_Properties,
                    ],
                    scope=SpaceIDScope(instance_spaces),
                )
            )
        if schema_spaces is not None:
            required_capabilities.append(
                DataModelsAcl(actions=[DataModelsAcl.Action.Read], scope=SpaceIDScope(schema_spaces)),
            )
        if missing := client.iam.verify_capabilities(required_capabilities):
            raise AuthenticationError(f"Missing required capabilities: {humanize_collection(missing)}.", missing)

    @staticmethod
    def validate_migration_model_available(client: ToolkitClient) -> None:
        models = client.data_modeling.data_models.retrieve([MODEL_ID], inline_views=False)
        if not models:
            raise ToolkitMigrationError(
                f"The migration data model {MODEL_ID!r} does not exist. "
                "Please run the `cdf migrate prepare` command to deploy the migration data model."
            )
        elif len(models) > 1:
            raise ToolkitMigrationError(
                f"Multiple migration models {MODEL_ID!r}. "
                "Please delete the duplicate models before proceeding with the migration."
            )
        model = models[0]
        missing_views = {INSTANCE_SOURCE_VIEW_ID, RESOURCE_VIEW_MAPPING_VIEW_ID} - set(model.views or [])
        if missing_views:
            raise ToolkitMigrationError(
                f"Invalid migration model. Missing views {humanize_collection(missing_views)}. "
                f"Please run the `cdf migrate prepare` command to deploy the migration data model."
            )



import re
import sys
import urllib
from contextlib import suppress
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, ClassVar

from rich import print

from cognite_toolkit import _version
from cognite_toolkit._cdf_tk.constants import clean_name
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitRequiredValueError,
    ToolkitTOMLFormatError,
    ToolkitVersionError,
)
from cognite_toolkit._cdf_tk.tk_warnings import MediumSeverityWarning

if sys.version_info >= (3, 11):
    from typing import Self

    import tomllib
    from tomllib import TOMLDecodeError
else:
    import tomli as tomllib
    from tomli import TOMLDecodeError
    from typing_extensions import Self


@dataclass
class CLIConfig:
    default_organization_dir: Path
    default_env: str = "dev"
    file_encoding: str | None = None
    has_user_set_default_org: bool = False
    has_user_set_default_env: bool = False

    @classmethod
    def load(cls, raw: dict[str, Any], cwd: Path) -> Self:
        has_user_set_default_org = "default_organization_dir" in raw
        has_user_set_default_env = "default_env" in raw
        default_organization_dir = cwd / raw["default_organization_dir"] if has_user_set_default_org else Path.cwd()
        return cls(
            default_organization_dir=default_organization_dir,
            default_env=raw.get("default_env", "dev"),
            file_encoding=raw.get("file_encoding"),
            has_user_set_default_org=has_user_set_default_org,
            has_user_set_default_env=has_user_set_default_env,
        )


@dataclass
class ModulesConfig:
    version: str
    packages: dict[str, list[str]] = field(default_factory=dict)

    @classmethod
    def load(cls, raw: dict[str, Any]) -> Self:
        version = raw["version"]
        packages = raw.get("packages", {})
        if (
            version != _version.__version__
            and (len(sys.argv) > 2 and sys.argv[1:3] != ["modules", "upgrade"])
            and "--help" not in sys.argv
            and version != "0.0.0"  # debugging mode
        ):
            raise ToolkitVersionError(
                f"The version of the modules ({version}) does not match the version of the installed CLI "
                f"({_version.__version__}). Please run `cdf modules upgrade` to upgrade the modules OR "
                f"run `pip install cognite-toolkit=={version}` to downgrade cdf CLI."
            )
        return cls(version=version, packages=packages)


@dataclass
class Library:
    url: str
    checksum: str

    def __post_init__(self) -> None:
        self._validate()

    def _validate(self) -> None:
        parsed_url = urllib.parse.urlparse(self.url)

        if not (parsed_url.scheme and parsed_url.netloc):
            raise ValueError("URL is missing scheme or network location (e.g., 'https://domain.com')")

        if parsed_url.scheme != "https":
            raise ValueError("URL must start with 'https'")

        if not parsed_url.path.casefold().endswith(".zip"):
            raise ValueError("URL must point to a .zip file.")

    @classmethod
    def load(cls, raw: dict[str, Any]) -> Self:
        if "url" not in raw:
            raise ValueError("Library configuration must contain 'url' field.")

        if "checksum" not in raw:
            raise ValueError("Library configuration must contain 'checksum' field.")

        return cls(**raw)


@dataclass
class CDFToml:
    """This is the configuration for the CLI and Modules"""

    file_name: ClassVar[str] = "cdf.toml"

    cdf: CLIConfig
    modules: ModulesConfig
    alpha_flags: dict[str, bool] = field(default_factory=dict)
    plugins: dict[str, bool] = field(default_factory=dict)
    libraries: dict[str, Library] = field(default_factory=dict)

    is_loaded_from_file: bool = False

    @classmethod
    def load(cls, cwd: Path | None = None, use_singleton: bool = True) -> "CDFToml":
        """Loads the cdf.toml file from the given path. If use_singleton is True, the instance will be stored as a
        singleton and returned on subsequent calls."""
        global _CDF_TOML
        if use_singleton and _CDF_TOML:
            return _CDF_TOML
        cwd = cwd or Path.cwd()
        file_path = cwd / cls.file_name
        if file_path.exists():
            raw = _read_toml(file_path)

            # No required fields in the cdf section
            cdf = CLIConfig.load(raw["cdf"], cwd) if "cdf" in raw else CLIConfig(cwd)
            try:
                modules = ModulesConfig.load(raw["modules"])
            except KeyError as e:
                raise ToolkitRequiredValueError(f"Missing required value in {cls.file_name}: {e.args}")
            alpha_flags = {}
            if "alpha_flags" in raw:
                alpha_flags = {clean_name(k): v for k, v in raw["alpha_flags"].items()}
            if not alpha_flags and "feature_flags" in raw:
                MediumSeverityWarning(
                    "The 'feature_flags' section has been renamed to 'alpha_flags'. Please update your cdf.toml file."
                ).print_warning()
                alpha_flags = {clean_name(k): v for k, v in raw["feature_flags"].items()}

            plugins = {}
            if "plugins" in raw:
                plugins = {clean_name(k): v for k, v in raw["plugins"].items()}

            libraries = {}
            for k, v in raw.get("library", {}).items():
                try:
                    libraries[k] = Library.load(v)
                except Exception as e:
                    raise ToolkitTOMLFormatError(f"Invalid library configuration for '{k}': {e.args[0]}") from e

            instance = cls(
                cdf=cdf,
                modules=modules,
                alpha_flags=alpha_flags,
                plugins=plugins,
                libraries=libraries,
                is_loaded_from_file=True,
            )
            if use_singleton:
                _CDF_TOML = instance
            return instance
        else:
            return cls(
                cdf=CLIConfig(cwd),
                modules=ModulesConfig.load({"version": _version.__version__}),
                alpha_flags={},
                plugins={},
                is_loaded_from_file=False,
            )


def _read_toml(file_path: Path) -> dict[str, Any]:
    # TOML files are required to be UTF-8 encoded
    content = file_path.read_text(encoding="utf-8")
    try:
        return tomllib.loads(content)
    except TOMLDecodeError as e:
        if file_path.is_relative_to(Path.cwd()):
            file_path = file_path.relative_to(Path.cwd())
        extra = ""
        with suppress(Exception):
            # If any errors is raised here, we ignore it and continue
            if result := re.search(r"\(at line (\d+), column (\d+)\)", str(e)):
                line_no = int(result.group(1))
                column_location = int(result.group(2))
                lines = content.splitlines()
                line = lines[line_no - 1]
                prefix = "\n    [blue]HINT: [/blue]"
                if column_location >= len(line):
                    extra = f"{prefix}This is near the end of the line '{line}'"
                elif column_location == 0:
                    extra = f"{prefix}This is near the beginning of the line '{line}'"
                else:
                    extra = f"{prefix}This is near the '{line[:column_location]}' in '{line[column_location:]}'"

        err = ToolkitTOMLFormatError(f"Error reading {file_path.as_posix()!r}: {e!s}{extra}")
        print(f"  [bold red]ERROR ([/][red]{type(err).__name__}[/][bold red]):[/] {err}")
        # We read the CDF.TML file from the module level, so the app has not been loaded yet.
        # Therefore, we raise SystemExit to stop the execution
        raise SystemExit(1)


_CDF_TOML: CDFToml | None = None

if __name__ == "__main__":
    from pprint import pprint

    # This is a test to quickly check that the code works.
    # also useful to check that when you change cdf.toml it is loaded correctly
    _ROOT = Path(__file__).parent.parent.parent
    pprint(CDFToml.load(_ROOT))  # noqa: T203



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes import FileMetadataYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_filemetadata_test_cases() -> Iterable:
    yield pytest.param(
        {"name": "MyFile"}, {"Missing required field: 'externalId'"}, id="Missing required field: externalId"
    )
    yield pytest.param(
        {"externalId": "my_file", "name": "TextFile", "mimeType": True},
        {
            "In field mimeType input should be a valid string. Got True of type bool. "
            "Hint: Use double quotes to force string."
        },
        id="Invalid mimeType type",
    )


class TestFileMetadataYAML:
    @pytest.mark.parametrize("data", list(find_resources("FileMetadata")))
    def test_load_valid_file_metadata(self, data: dict[str, object]) -> None:
        loaded = FileMetadataYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_filemetadata_test_cases()))
    def test_invalid_asset_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, FileMetadataYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



from pydantic import Field, JsonValue

from cognite_toolkit._cdf_tk.resource_classes.base import ToolkitResource


class RobotDataPostProcessingYAML(ToolkitResource):
    name: str = Field(description="DataProcessing name.")
    external_id: str = Field(
        description="DataProcessing external id. Must be unique for the resource type.",
    )
    method: str = Field(description="DataProcessing method, to call the right functionality on the robot.")
    input_schema: dict[str, JsonValue] | None = Field(
        default=None, description="Schema that defines what inputs are needed for the data postprocessing."
    )
    description: str | None = Field(default=None, description="Textual description of the DataProcessing.")



from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any
from unittest.mock import MagicMock

from cognite.client._api.datapoints import DatapointsAPI
from cognite.client._api.datapoints_subscriptions import DatapointsSubscriptionAPI
from cognite.client._api.functions import FunctionCallsAPI, FunctionSchedulesAPI
from cognite.client._api.raw import RawDatabasesAPI, RawRowsAPI, RawTablesAPI
from cognite.client._api.synthetic_time_series import SyntheticDatapointsAPI
from cognite.client.testing import CogniteClientMock
from rich.console import Console

from cognite_toolkit._cdf_tk.client._toolkit_client import ToolkitClient

from .api.canvas import CanvasAPI, IndustrialCanvasAPI
from .api.charts import ChartsAPI
from .api.dml import DMLAPI
from .api.extended_data_modeling import ExtendedInstancesAPI
from .api.extended_files import ExtendedFileMetadataAPI
from .api.extended_functions import ExtendedFunctionsAPI
from .api.extended_raw import ExtendedRawAPI
from .api.extended_timeseries import ExtendedTimeSeriesAPI
from .api.infield import InfieldAPI, InfieldConfigAPI
from .api.location_filters import LocationFiltersAPI
from .api.lookup import (
    AssetLookUpAPI,
    DataSetLookUpAPI,
    EventLookUpAPI,
    ExtractionPipelineLookUpAPI,
    FileMetadataLookUpAPI,
    FunctionLookUpAPI,
    LocationFiltersLookUpAPI,
    LookUpGroup,
    SecurityCategoriesLookUpAPI,
    TimeSeriesLookUpAPI,
)
from .api.migration import CreatedSourceSystemAPI, InstanceSourceAPI, MigrationAPI, ResourceViewMappingAPI
from .api.project import ProjectAPI
from .api.robotics import RoboticsAPI
from .api.robotics.capabilities import CapabilitiesAPI
from .api.robotics.data_postprocessing import DataPostProcessingAPI
from .api.robotics.frames import FramesAPI
from .api.robotics.locations import LocationsAPI as RoboticsLocationsAPI
from .api.robotics.maps import MapsAPI
from .api.search import SearchAPI
from .api.search_config import SearchConfigurationsAPI
from .api.streams import StreamsAPI
from .api.token import TokenAPI
from .api.verify import VerifyAPI


class ToolkitClientMock(CogniteClientMock):
    """Mock for ToolkitClient object

    All APIs are replaced with specked MagicMock objects.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if "parent" in kwargs:
            super().__init__(*args, **kwargs)
            return None
        super().__init__(*args, **kwargs)
        self.console = Console()
        # Developer note:
        # - Please add your mocked APIs in chronological order
        # - For nested APIs:
        #   - Add spacing above and below
        #   - Use `spec=MyAPI` only for "top level"
        #   - Use `spec_set=MyNestedAPI` for all nested APIs
        self.canvas = MagicMock(spec=CanvasAPI)
        self.canvas.industrial = MagicMock(spec_set=IndustrialCanvasAPI)
        self.charts = MagicMock(spec_set=ChartsAPI)
        self.files = MagicMock(spec_set=ExtendedFileMetadataAPI)
        self.functions = MagicMock(spec=ExtendedFunctionsAPI)
        self.functions.calls = MagicMock(spec_set=FunctionCallsAPI)
        self.functions.schedules = MagicMock(spec_set=FunctionSchedulesAPI)
        self.infield = MagicMock(spec=InfieldAPI)
        self.infield.config = MagicMock(spec_set=InfieldConfigAPI)

        self.project = MagicMock(spec_set=ProjectAPI)

        self.search = MagicMock(spec=SearchAPI)
        self.search.locations = MagicMock(spec_set=LocationFiltersAPI)
        self.search.configurations = MagicMock(spec_set=SearchConfigurationsAPI)
        self.dml = MagicMock(spec_set=DMLAPI)
        self.lookup = MagicMock(spec=LookUpGroup)
        self.lookup.data_sets = MagicMock(spec_set=DataSetLookUpAPI)
        self.lookup.assets = MagicMock(spec_set=AssetLookUpAPI)
        self.lookup.time_series = MagicMock(spec_set=TimeSeriesLookUpAPI)
        self.lookup.files = MagicMock(spec_set=FileMetadataLookUpAPI)
        self.lookup.events = MagicMock(spec_set=EventLookUpAPI)
        self.lookup.security_categories = MagicMock(spec_set=SecurityCategoriesLookUpAPI)
        self.lookup.location_filters = MagicMock(spec_set=LocationFiltersLookUpAPI)
        self.lookup.extraction_pipelines = MagicMock(spec_set=ExtractionPipelineLookUpAPI)
        self.lookup.functions = MagicMock(spec_set=FunctionLookUpAPI)
        self.migration = MagicMock(spec=MigrationAPI)
        self.migration.instance_source = MagicMock(spec_set=InstanceSourceAPI)
        self.migration.resource_view_mapping = MagicMock(spec_set=ResourceViewMappingAPI)
        self.migration.created_source_system = MagicMock(spec_set=CreatedSourceSystemAPI)
        self.raw = MagicMock(spec=ExtendedRawAPI)
        self.raw.databases = MagicMock(spec_set=RawDatabasesAPI)
        self.raw.rows = MagicMock(spec_set=RawRowsAPI)
        self.raw.tables = MagicMock(spec_set=RawTablesAPI)

        self.robotics = MagicMock()
        self.robotics.robots = MagicMock(spec=RoboticsAPI)
        self.robotics.data_postprocessing = MagicMock(spec_set=DataPostProcessingAPI)
        self.robotics.locations = MagicMock(spec_set=RoboticsLocationsAPI)
        self.robotics.frames = MagicMock(spec_set=FramesAPI)
        self.robotics.maps = MagicMock(spec_set=MapsAPI)
        self.robotics.capabilities = MagicMock(spec_set=CapabilitiesAPI)

        self.data_modeling.instances = MagicMock(spec_set=ExtendedInstancesAPI)

        self.time_series = MagicMock(spec=ExtendedTimeSeriesAPI)
        self.time_series.data = MagicMock(spec=DatapointsAPI)
        self.time_series.data.synthetic = MagicMock(spec_set=SyntheticDatapointsAPI)
        self.time_series.subscriptions = MagicMock(spec_set=DatapointsSubscriptionAPI)

        self.streams = MagicMock(spec=StreamsAPI)

        # This is a helper API, not a real API.
        self.token = TokenAPI(self)
        self.verify = MagicMock(spec_set=VerifyAPI)


@contextmanager
def monkeypatch_toolkit_client() -> Iterator[ToolkitClientMock]:
    toolkit_client_mock = ToolkitClientMock()
    try:
        ToolkitClient.__new__ = lambda *args, **kwargs: toolkit_client_mock  # type: ignore[method-assign]
        yield toolkit_client_mock
    finally:
        ToolkitClient.__new__ = lambda cls, *args, **kwargs: object.__new__(cls)  # type: ignore[method-assign]



from cognite.client.data_classes import DataSet

from cognite_toolkit._cdf_tk.cruds import DataSetsCRUD, ResourceWorker
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from tests.data import LOAD_DATA
from tests.test_unit.approval_client import ApprovalToolkitClient


class TestDataSetsLoader:
    def test_upsert_data_set(
        self, env_vars_with_client: EnvironmentVariables, toolkit_client_approval: ApprovalToolkitClient
    ):
        loader = DataSetsCRUD.create_loader(env_vars_with_client.get_client())
        raw_list = loader.load_resource_file(
            LOAD_DATA / "data_sets" / "1.my_datasets.yaml", env_vars_with_client.dump()
        )
        assert len(raw_list) == 2

        first = DataSet._load(raw_list[0])
        # Set the properties that are set on the server side
        first.id = 42
        first.created_time = 42
        first.last_updated_time = 42
        # Simulate that the data set is already in CDF
        toolkit_client_approval.append(DataSet, first)

        worker = ResourceWorker(loader, "deploy")
        resources = worker.prepare_resources([LOAD_DATA / "data_sets" / "1.my_datasets.yaml"])

        assert {
            "create": len(resources.to_create),
            "change": len(resources.to_update),
            "delete": len(resources.to_delete),
            "unchanged": len(resources.unchanged),
        } == {"create": 1, "change": 0, "delete": 0, "unchanged": 1}



import json

import pytest
import responses
from cognite.client.data_classes.data_modeling import (
    Edge,
    EdgeApply,
    EdgeId,
    Node,
    NodeApply,
    NodeId,
    NodeOrEdgeData,
    ViewId,
)
from cognite.client.data_classes.data_modeling.cdm.v1 import (
    CogniteAnnotationApply,
    CogniteAssetApply,
    CogniteTimeSeriesApply,
)
from cognite.client.exceptions import CogniteAPIError, CogniteConnectionError, CogniteReadTimeout

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.instances import InstanceList


@pytest.fixture()
def some_timeseries() -> CogniteTimeSeriesApply:
    return CogniteTimeSeriesApply(
        space="some_space",
        external_id="test_time_series",
        is_step=False,
        time_series_type="numeric",
    )


class TestInstances:
    def test_apply_fast_failed_multiple_types(self, toolkit_config) -> None:
        different_instance_types = [
            CogniteTimeSeriesApply(
                space="some_space",
                external_id="test_time_series",
                is_step=False,
                time_series_type="numeric",
            ),
            CogniteAssetApply(
                space="some_space",
                external_id="test_asset",
                name="Test Asset",
            ),
            CogniteAnnotationApply(
                space="some_space",
                external_id="test_annotation",
                type=("some_space", "entity.match"),
                start_node=("some_space", "test_asset"),
                end_node=("some_space", "test_asset"),
                name="Test Annotation",
            ),
        ]
        url = f"{toolkit_config.base_url}/api/v1/projects/test-project/models/instances"
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.POST,
                url,
                status=400,
                json={"error": "Invalid request", "message": "the message"},
            )
            client = ToolkitClient(config=toolkit_config)
            with pytest.raises(CogniteAPIError) as exc_info:
                _ = client.data_modeling.instances.apply_fast(different_instance_types)

        error = exc_info.value
        assert isinstance(error, CogniteAPIError)
        assert error.code == 400
        assert error.message == "Invalid request"
        assert error.failed == [instance.as_id() for instance in different_instance_types]

    @pytest.mark.usefixtures("disable_gzip")
    def test_apply_fast_429_status_split(self, toolkit_config: ToolkitClientConfig) -> None:
        instances = [
            CogniteTimeSeriesApply(
                space="some_space",
                external_id=f"test_time_series_{i}",
                is_step=False,
                time_series_type="numeric",
            )
            for i in range(2)
        ]
        url = f"{toolkit_config.base_url}/api/v1/projects/test-project/models/instances"
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.POST,
                url,
                status=429,
                json={"error": "Too many items"},
            )
            for ts in instances:
                rsps.add(
                    responses.POST,
                    url,
                    status=200,
                    json={
                        "items": [
                            {
                                "instanceType": "node",
                                "space": ts.space,
                                "externalId": ts.external_id,
                                "version": 1,
                                "wasModified": False,
                                "createdTime": 0,
                                "lastUpdatedTime": 0,
                            }
                        ]
                    },
                )
            client = ToolkitClient(config=toolkit_config)

            results = client.data_modeling.instances.apply_fast(instances)

            assert len(results) == 2
            assert len(rsps.calls) >= 3
            failed_call = rsps.calls[-3]
            assert failed_call.response.status_code == 429
            assert len(json.loads(failed_call.request.body)["items"]) == 2
            for call in rsps.calls[-2:]:
                assert call.response.status_code == 200
                assert len(json.loads(call.request.body)["items"]) == 1

    @pytest.mark.usefixtures("disable_gzip")
    def test_apply_fast_429_single_instance(
        self, toolkit_config: ToolkitClientConfig, some_timeseries: CogniteTimeSeriesApply
    ) -> None:
        url = f"{toolkit_config.base_url}/api/v1/projects/test-project/models/instances"
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.POST,
                url,
                status=429,
                json={"error": "Too many items"},
            )
            rsps.add(
                responses.POST,
                url,
                status=200,
                json={
                    "items": [
                        {
                            "instanceType": "node",
                            "space": some_timeseries.space,
                            "externalId": some_timeseries.external_id,
                            "version": 1,
                            "wasModified": False,
                            "createdTime": 0,
                            "lastUpdatedTime": 0,
                        }
                    ]
                },
            )
            client = ToolkitClient(config=toolkit_config)

            results = client.data_modeling.instances.apply_fast([some_timeseries])

            assert len(results) == 1
            assert len(rsps.calls) >= 2
            failed_call = rsps.calls[-2]
            assert failed_call.response.status_code == 429
            assert len(json.loads(failed_call.request.body)["items"]) == 1
            last_call = rsps.calls[-1]
            assert last_call.response.status_code == 200
            assert len(json.loads(last_call.request.body)["items"]) == 1

    @pytest.mark.usefixtures("max_retries_2")
    @pytest.mark.parametrize(
        "args, expected_exception",
        [
            pytest.param(
                dict(status=429, json={"error": "Server is busy, please try again later"}),
                CogniteAPIError,
                id="429 Too Many Requests",
            ),
            pytest.param(
                dict(body=ConnectionRefusedError("Connection refused")), CogniteConnectionError, id="Connection refused"
            ),
            pytest.param(dict(body=TimeoutError("timed out")), CogniteReadTimeout, id="Connection timed out"),
        ],
    )
    def test_apply_fast_raise(
        self,
        args: dict[str, object],
        expected_exception: type[Exception],
        toolkit_config: ToolkitClientConfig,
        some_timeseries: CogniteTimeSeriesApply,
    ) -> None:
        url = f"{toolkit_config.base_url}/api/v1/projects/test-project/models/instances"
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.POST,
                url,
                **args,
            )
            client = ToolkitClient(config=toolkit_config)
            with pytest.raises(expected_exception) as _:
                _ = client.data_modeling.instances.apply_fast([some_timeseries])

    def test_apply_fast_invalid_json(self, toolkit_config: ToolkitClientConfig) -> None:
        node = NodeApply(
            space="some_space",
            external_id="test_time_series",
            sources=[
                NodeOrEdgeData(source=ViewId("some_space", "some_view", "v1"), properties={"my_number": float("nan")})
            ],
        )
        client = ToolkitClient(config=toolkit_config)

        with pytest.raises(ValueError) as exc_info:
            _ = client.data_modeling.instances.apply_fast([node])

        assert "Out of range float values are not JSON compliant." in str(exc_info.value)

    def test_apply_fast_empty(self, toolkit_config: ToolkitClientConfig) -> None:
        client = ToolkitClient(config=toolkit_config)
        result = client.data_modeling.instances.apply_fast([])
        assert len(result) == 0

    @pytest.mark.usefixtures("disable_gzip")
    def test_delete_fast_400(self, toolkit_config: ToolkitClientConfig) -> None:
        url = f"{toolkit_config.base_url}/api/v1/projects/test-project/models/instances/delete"
        error_message = "This will trigger a 400 error. 287891n-unique"
        ids = [NodeId("node-space", "node1"), EdgeId("edge-space", "edge1")]
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.POST,
                url,
                status=400,
                json={"error": error_message},
            )
            client = ToolkitClient(config=toolkit_config)

            with pytest.raises(CogniteAPIError) as exc_info:
                _ = client.data_modeling.instances.delete_fast(ids)

            error = exc_info.value
            assert isinstance(error, CogniteAPIError)
            assert error.code == 400
            assert error.message == error_message
            assert error.failed == ids


class TestInstancesDataClasses:
    def test_instance_list_mixed_instance_types(self) -> None:
        node = Node(
            space="some_space",
            external_id="test_node",
            version=1,
            created_time=0,
            last_updated_time=0,
            deleted_time=None,
            type=None,
            properties=None,
        )
        edge = Edge(
            space="some_space",
            external_id="test_edge",
            version=1,
            created_time=0,
            last_updated_time=0,
            deleted_time=None,
            type=("some_space", "some_relation"),
            properties=None,
            start_node=("some_space", "start_node"),
            end_node=("some_space", "end_node"),
        )
        my_list = InstanceList([node, edge])
        assert len(my_list) == 2
        my_write_list = my_list.as_write()
        assert len(my_write_list) == 2
        assert isinstance(my_write_list[0], NodeApply)
        assert isinstance(my_write_list[1], EdgeApply)



import pytest
from cognite.client.data_classes import DataPointSubscriptionWrite
from cognite.client.data_classes.data_modeling import NodeId
from cognite.client.data_classes.datapoints_subscriptions import TimeSeriesID, TimeSeriesIDList

from cognite_toolkit._cdf_tk.cruds import DatapointSubscriptionCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError


class TestDatapointSubscriptionLoader:
    @pytest.mark.parametrize(
        "ts_count, node_count, expected_ts_count, expected_node_count, expected_updates",
        [
            pytest.param(150, 150, 100, 0, 2, id="150 timeseries, 150 nodes"),
            pytest.param(20, 0, 20, 0, 0, id="20 timeseries, no nodes"),
            pytest.param(101, 0, 100, 0, 1, id="101 timeseries, no nodes"),
            pytest.param(0, 20, 0, 20, 0, id="no timeseries, 20 nodes"),
            pytest.param(0, 101, 0, 100, 1, id="no timeseries, 101 nodes"),
            pytest.param(50, 150, 50, 50, 1, id="50 timeseries, 150 nodes"),
            pytest.param(150, 50, 100, 0, 1, id="150 timeseries, 50 nodes"),
        ],
    )
    def test_create_split_timeseries_ids(
        self, ts_count: int, node_count: int, expected_ts_count: int, expected_node_count: int, expected_updates: int
    ) -> None:
        sub = DataPointSubscriptionWrite(
            external_id="mySub",
            partition_count=1,
            time_series_ids=[f"timeseries_{i}" for i in range(ts_count)] or None,
            instance_ids=[NodeId("my_space", f"node_{i}") for i in range(node_count)] or None,
        )

        to_upsert, batches = DatapointSubscriptionCRUD.create_split_timeseries_ids(sub)

        assert len(to_upsert.time_series_ids or []) == expected_ts_count
        assert len(to_upsert.instance_ids or []) == expected_node_count
        assert len(batches) == expected_updates

    def test_create_split_timeseries_ids_raise(self) -> None:
        sub = DataPointSubscriptionWrite(
            external_id="mySub",
            partition_count=1,
            time_series_ids=[f"timeseries_{i}" for i in range(6_000)],
            instance_ids=[NodeId("my_space", f"node_{i}") for i in range(6_000)],
        )

        with pytest.raises(ToolkitValueError) as exc:
            DatapointSubscriptionCRUD.create_split_timeseries_ids(sub)

        assert str(exc.value) == 'Subscription "mySub" has 12,000 time series, which is more than the limit of 10,000.'

    @pytest.mark.parametrize(
        "ts_count, node_count, existing_ts_count, existing_node_count, expected_updates",
        [
            pytest.param(150, 150, 100, 0, 2, id="Adding 50 timeseries and 150 nodes"),
            pytest.param(20, 0, 20, 0, 0, id="No change with existing 20 timeseries"),
            pytest.param(101, 0, 100, 0, 1, id="Adding 1 timeseries to existing 100"),
            pytest.param(0, 20, 0, 20, 0, id="No change with existing nodes"),
            pytest.param(202, 303, 403, 0, 4, id="Removing 101 timeseries and adding 303 nodes"),
            pytest.param(1, 1, 400, 600, 10, id="Removing 399 timeseries and adding 601 nodes"),
        ],
    )
    def test_update_split_timeseries_ids(
        self, ts_count: int, node_count: int, existing_ts_count: int, existing_node_count: int, expected_updates: int
    ) -> None:
        current = TimeSeriesIDList([])
        if existing_ts_count > 0:
            current.extend([TimeSeriesID(external_id=f"timeseries_{i}", id=i) for i in range(existing_ts_count)])
        if existing_node_count > 0:
            current.extend(
                [
                    TimeSeriesID(instance_id=NodeId("my_space", f"node_{i}"), id=i + existing_ts_count)
                    for i in range(existing_node_count)
                ]
            )
        sub = DataPointSubscriptionWrite(
            external_id="mySub",
            partition_count=1,
            time_series_ids=[f"timeseries_{i}" for i in range(ts_count)] or None,
            instance_ids=[NodeId("my_space", f"node_{i}") for i in range(node_count)] or None,
        )

        _, batches = DatapointSubscriptionCRUD.update_split_timeseries_ids(sub, current)
        assert len(batches) == expected_updates

    def test_update_split_timeseries_ids_raise(self) -> None:
        current = TimeSeriesIDList([TimeSeriesID(external_id=f"timeseries_{i}", id=i) for i in range(6_000)])
        sub = DataPointSubscriptionWrite(
            external_id="mySub",
            partition_count=1,
            time_series_ids=[f"timeseries_{i}" for i in range(6_000)],
            instance_ids=[NodeId("my_space", f"node_{i}") for i in range(6_000)],
        )

        with pytest.raises(ToolkitValueError) as exc:
            DatapointSubscriptionCRUD.update_split_timeseries_ids(sub, current)

        assert str(exc.value) == 'Subscription "mySub" has 12,000 time series, which is more than the limit of 10,000.'



from cognite.client import CogniteClient
from cognite.client._api_client import APIClient
from cognite.client.config import ClientConfig

from .location_filters import LocationFiltersAPI
from .search_config import SearchConfigurationsAPI


class SearchAPI(APIClient):
    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: CogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        self.locations = LocationFiltersAPI(config, api_version, cognite_client)
        self.configurations = SearchConfigurationsAPI(config, api_version, cognite_client)



from pathlib import Path
from unittest.mock import MagicMock

import pytest
from cognite.client.data_classes import Asset
from cognite.client.data_classes.data_modeling import (
    ContainerId,
    DirectRelation,
    DirectRelationReference,
    InstanceApply,
    MappedProperty,
    NodeId,
    NodeList,
    Text,
    View,
    ViewId,
)

from cognite_toolkit._cdf_tk.client.data_classes.migration import CreatedSourceSystem, ResourceViewMapping
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.commands._migrate.data_classes import (
    AssetCentricMapping,
    AssetCentricMappingList,
    MigrationMapping,
)
from cognite_toolkit._cdf_tk.commands._migrate.data_mapper import AssetCentricMapper
from cognite_toolkit._cdf_tk.commands._migrate.issues import ConversionIssue, MigrationIssue
from cognite_toolkit._cdf_tk.commands._migrate.selectors import MigrationCSVFileSelector
from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError


class TestAssetCentricMapper:
    def test_map_assets(self, tmp_path: Path) -> None:
        asset_count = 10
        source = AssetCentricMappingList(
            [
                AssetCentricMapping(
                    mapping=MigrationMapping(
                        resourceType="asset",
                        instanceId=NodeId(space="my_space", external_id=f"asset_{i}"),
                        id=1000 + i,
                        ingestionView="cdf_asset_mapping",
                    ),
                    resource=Asset(
                        id=1000 + i,
                        name=f"Asset {i}",
                        source="sap",
                        # Half of the assets will be missing description and thus have a conversion issue.
                        description=f"Description {i}" if i % 2 == 0 else None,
                    ),
                )
                for i in range(asset_count)
            ]
        )
        mapping_file = tmp_path / "mapping.csv"
        mapping_file.write_text(
            "id,space,externalId,ingestionView\n"
            + "\n".join(f"{1000 + i},my_space,asset_{i},cdf_asset_mapping" for i in range(asset_count))
        )

        selected = MigrationCSVFileSelector(datafile=mapping_file, kind="Assets")

        with monkeypatch_toolkit_client() as client:
            client.migration.resource_view_mapping.retrieve.return_value = NodeList[ResourceViewMapping](
                [
                    ResourceViewMapping(
                        external_id="cdf_asset_mapping",
                        resource_type="asset",
                        view_id=ViewId("cdf_cdm", "CogniteAsset", "v1"),
                        property_mapping={
                            "name": "name",
                            "description": "description",
                            "source": "source",
                        },
                        last_updated_time=1,
                        created_time=0,
                        version=1,
                    )
                ]
            )
            client.migration.created_source_system.list.return_value = NodeList[CreatedSourceSystem](
                [
                    CreatedSourceSystem(
                        space="source_systems",
                        external_id="SAP",
                        source="sap",
                        last_updated_time=1,
                        created_time=0,
                        version=1,
                    ),
                ]
            )
            # Mocking the view to avoid setting all properties we don't use
            cognite_asset = MagicMock(spec=View)
            cognite_asset.properties = {
                "name": MappedProperty(
                    ContainerId("cdf_cdm", "CogniteDescribable"), "name", Text(), True, False, False
                ),
                "description": MappedProperty(
                    ContainerId("cdf_cdm", "CogniteDescribable"), "description", Text(), True, False, False
                ),
                "source": MappedProperty(
                    ContainerId("cdf_cdm", "CogniteSourceable"), "source", DirectRelation(False), True, False, False
                ),
            }
            cognite_asset.as_id.return_value = ViewId("cdf_cdm", "CogniteAsset", "v1")
            client.data_modeling.views.retrieve.return_value = [cognite_asset]

            mapper = AssetCentricMapper(client)

            mapper.prepare(selected)

            mapped: list[InstanceApply] = []
            issues: list[MigrationIssue] = []
            for item in source:
                target, item_issue = mapper.map(item)
                mapped.append(target)
                if not isinstance(item_issue, ConversionIssue) or item_issue.has_issues:
                    issues.append(item_issue)

            # We do not assert the exact content of mapped, as that is tested in the
            # tests for the asset_centric_to_dm function.
            assert len(mapped) == asset_count
            assert len(issues) == asset_count // 2
            # All issues are the same.
            first_issue = issues[0]
            assert isinstance(first_issue, ConversionIssue)
            assert first_issue.missing_asset_centric_properties == ["description"]
            first_asset = mapped[0]
            assert first_asset.sources[0].properties["source"] == DirectRelationReference("source_systems", "SAP")

            assert client.migration.resource_view_mapping.retrieve.call_count == 1
            client.migration.resource_view_mapping.retrieve.assert_called_with(["cdf_asset_mapping"])
            assert client.migration.created_source_system.list.call_count == 1
            assert client.data_modeling.views.retrieve.call_count == 1
            client.data_modeling.views.retrieve.assert_called_with([ViewId("cdf_cdm", "CogniteAsset", "v1")])

    def test_map_chunk_before_prepare_raises_error(self, tmp_path: Path) -> None:
        """Test that calling map_chunk before prepare raises a RuntimeError."""
        source = AssetCentricMapping(
            mapping=MigrationMapping(
                resourceType="asset",
                instanceId=NodeId(space="my_space", external_id="asset_1"),
                id=1001,
                ingestionView="cdf_asset_mapping",
            ),
            resource=Asset(
                id=1001,
                name="Asset 1",
                description="Description 1",
            ),
        )

        with monkeypatch_toolkit_client() as client:
            mapper = AssetCentricMapper(client)

            # Call map_chunk without calling prepare first
            with pytest.raises(
                RuntimeError,
                match=r"Failed to lookup mapping or view for ingestion view 'cdf_asset_mapping'. Did you forget to call .prepare()?",
            ):
                mapper.map(source)

    def test_prepare_missing_view_source_raises_error(self, tmp_path: Path) -> None:
        """Test that prepare raises ToolkitValueError when view source is not found."""
        mapping_file = tmp_path / "mapping.csv"
        mapping_file.write_text("id,space,externalId,ingestionView\n1001,my_space,asset_1,missing_view_source")

        selected = MigrationCSVFileSelector(datafile=mapping_file, kind="Assets")

        with monkeypatch_toolkit_client() as client:
            # Return empty list to simulate missing view source
            client.migration.resource_view_mapping.retrieve.return_value = NodeList[ResourceViewMapping]([])

            mapper = AssetCentricMapper(client)

            with pytest.raises(
                ToolkitValueError, match=r"The following ingestion views were not found: missing_view_source"
            ):
                mapper.prepare(selected)

    def test_prepare_missing_view_in_data_modeling_raises_error(self, tmp_path: Path) -> None:
        """Test that prepare raises ToolkitValueError when view is not found in Data Modeling."""
        mapping_file = tmp_path / "mapping.csv"
        mapping_file.write_text("id,space,externalId,ingestionView\n1001,my_space,asset_1,cdf_asset_mapping")

        selected = MigrationCSVFileSelector(datafile=mapping_file, kind="Assets")

        with monkeypatch_toolkit_client() as client:
            # Return view source but empty view list to simulate missing view in Data Modeling
            client.migration.resource_view_mapping.retrieve.return_value = NodeList[ResourceViewMapping](
                [
                    ResourceViewMapping(
                        external_id="cdf_asset_mapping",
                        resource_type="asset",
                        view_id=ViewId("cdf_cdm", "CogniteAsset", "v1"),
                        property_mapping={
                            "name": "name",
                            "description": "description",
                        },
                        last_updated_time=1,
                        created_time=0,
                        version=1,
                    )
                ]
            )
            # Return empty list to simulate missing view in Data Modeling
            client.data_modeling.views.retrieve.return_value = []

            mapper = AssetCentricMapper(client)

            with pytest.raises(ToolkitValueError) as exc_info:
                mapper.prepare(selected)

            assert (
                str(exc_info.value)
                == "The following ingestion views were not found in Data Modeling: ViewId(space='cdf_cdm', external_id='CogniteAsset', version='v1')"
            )



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.hosted_extractor_mapping import HostedExtractorMappingYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_hosted_extractor_mapping_test_cases() -> Iterable:
    yield pytest.param(
        {"externalId": "myMapping", "mapping": "var->to->other", "published": True},
        {
            "In field mapping input must be an object of type Mapping. Got 'var->to->other' of type str.",
        },
        id="Incorrect variables type ",
    )
    yield pytest.param(
        {
            "externalId": "myMapping",
            "mapping": {"expression": "some_expression"},
            "published": True,
            "input": {
                "type": "csv",
                "delimiter": ":,",
            },
        },
        {"In input.delimiter string should have at most 1 character"},
        id="Invalid delimiter in CSV input",
    )


class TestHostedExtractorMappingYAML:
    @pytest.mark.parametrize("data", list(find_resources("Mapping", resource_dir="hosted_extractors")))
    def test_load_valid_hosted_extractor_mapping(self, data: dict[str, object]) -> None:
        loaded = HostedExtractorMappingYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_hosted_extractor_mapping_test_cases()))
    def test_invalid_hosted_extractor_mapping_error_messages(
        self, data: dict | list, expected_errors: set[str]
    ) -> None:
        warning_list = validate_resource_yaml_pydantic(data, HostedExtractorMappingYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



from pathlib import Path
from typing import Annotated

import typer
from rich import print

from cognite_toolkit._cdf_tk.commands import RepoCommand
from cognite_toolkit._cdf_tk.commands.repo import REPOSITORY_HOSTING
from cognite_toolkit._cdf_tk.utils import humanize_collection


class RepoApp(typer.Typer):
    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.main)
        self.command()(self.init)

    def main(self, ctx: typer.Context) -> None:
        """Commands to repo management"""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf repo --help[/] for more information.")

    def init(
        self,
        cwd: Annotated[
            Path,
            typer.Argument(
                help="",
            ),
        ] = Path.cwd(),
        host: str | None = typer.Option(
            None,
            "--host",
            "-h",
            help=f"Hosting service for the repository. Supported {humanize_collection(REPOSITORY_HOSTING)}. If not provided, you will be prompted to choose.",
        ),
        verbose: bool = typer.Option(False, "-v", "--verbose", help="Verbose output"),
    ) -> None:
        """Initialize a new git repository with files like .gitignore, cdf.toml, and so on."""

        cmd = RepoCommand()
        cmd.run(lambda: cmd.init(cwd=cwd, host=host, verbose=verbose))



import warnings
from collections.abc import Sequence
from itertools import groupby
from typing import TypeVar, overload

from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes.data_modeling import (
    Node,
    NodeApplyResult,
    NodeApplyResultList,
    NodeId,
    NodeList,
    filters,
    query,
)
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client.data_classes.migration import (
    AssetCentricId,
    CreatedSourceSystem,
    InstanceSource,
    ResourceViewMapping,
    ResourceViewMappingApply,
    SpaceSource,
)
from cognite_toolkit._cdf_tk.constants import COGNITE_MIGRATION_SPACE
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence
from cognite_toolkit._cdf_tk.utils.useful_types import AssetCentricType

from .extended_data_modeling import ExtendedInstancesAPI


class InstanceSourceAPI:
    def __init__(self, instance_api: ExtendedInstancesAPI) -> None:
        self._instance_api = instance_api
        self._RETRIEVE_LIMIT = 1000
        self._view_id = InstanceSource.get_source()

    def retrieve(self, ids: Sequence[AssetCentricId]) -> NodeList[InstanceSource]:
        """Retrieve a list of instance sources by their IDs.

        Args:
            ids (Sequence[AssetCentricId]): A sequence of AssetCentricId objects representing the IDs of the instance sources to retrieve.
        """
        results: NodeList[InstanceSource] = NodeList[InstanceSource]([])
        for chunk in chunker_sequence(ids, self._RETRIEVE_LIMIT):
            retrieve_query = query.Query(
                with_={
                    "instanceSource": query.NodeResultSetExpression(
                        filter=filters.And(filters.HasData(views=[self._view_id]), self._create_dms_filter(ids)),
                        limit=len(chunk),
                    ),
                },
                select={"instanceSource": query.Select([query.SourceSelector(self._view_id, ["*"])])},
            )
            chunk_response = self._instance_api.query(retrieve_query)
            results.extend([InstanceSource._load(item.dump()) for item in chunk_response.get("instanceSource", [])])
        return results

    @staticmethod
    def _create_dms_filter(ids: Sequence[AssetCentricId]) -> filters.Filter:
        """Create a filter that matches all the AssetCentricIds in the list."""
        if not ids:
            raise ValueError("Cannot create a filter from an empty AssetCentricIdList.")
        to_or_filters: list[filters.Filter] = []
        instance_source_view = InstanceSource.get_source()
        for resource_type, resource_ids in groupby(
            sorted(ids, key=lambda x: x.resource_type), key=lambda x: x.resource_type
        ):
            is_resource = filters.Equals(instance_source_view.as_property_ref("resourceType"), resource_type)
            is_id = filters.In(
                instance_source_view.as_property_ref("id"), [resource_id.id_ for resource_id in resource_ids]
            )
            to_or_filters.append(filters.And(is_resource, is_id))
        return filters.Or(*to_or_filters)

    def list(self, resource_type: AssetCentricType, limit: int | None = DEFAULT_LIMIT_READ) -> NodeList[InstanceSource]:
        """List instance sources filtered by resource type"""
        is_selected = filters.Equals(self._view_id.as_property_ref("resourceType"), resource_type)

        nodes = self._instance_api.list(instance_type="node", filter=is_selected, limit=limit, sources=self._view_id)
        return NodeList[InstanceSource]([InstanceSource._load(node.dump()) for node in nodes])


class ResourceViewMappingAPI:
    """API for retrieving resource view mapping from the data model."""

    def __init__(self, instance_api: ExtendedInstancesAPI) -> None:
        self._instance_api = instance_api
        self._view_id = ResourceViewMapping.get_source()

    @overload
    def upsert(
        self,
        item: ResourceViewMappingApply,
        skip_on_version_conflict: bool = False,
        replace: bool = False,
    ) -> NodeApplyResult: ...

    @overload
    def upsert(
        self,
        item: Sequence[ResourceViewMappingApply],
        skip_on_version_conflict: bool = False,
        replace: bool = False,
    ) -> NodeApplyResultList: ...

    def upsert(
        self,
        item: ResourceViewMappingApply | Sequence[ResourceViewMappingApply],
        skip_on_version_conflict: bool = False,
        replace: bool = False,
    ) -> NodeApplyResult | NodeApplyResultList:
        """Upsert one or more view sources."""
        result = self._instance_api.apply(
            item, skip_on_version_conflict=skip_on_version_conflict, replace=replace
        ).nodes
        if isinstance(item, ResourceViewMappingApply):
            return result[0]
        return NodeApplyResultList(result)

    @overload
    def retrieve(self, external_id: str) -> ResourceViewMapping | None: ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str]) -> NodeList[ResourceViewMapping]: ...

    def retrieve(
        self, external_id: str | SequenceNotStr[str]
    ) -> ResourceViewMapping | NodeList[ResourceViewMapping] | None:
        """Retrieve one or more view sources by their external IDs."""
        if isinstance(external_id, str):
            return self._instance_api.retrieve_nodes(
                NodeId(COGNITE_MIGRATION_SPACE, external_id), node_cls=ResourceViewMapping
            )
        else:
            nodes = self._instance_api.retrieve(
                nodes=[NodeId(COGNITE_MIGRATION_SPACE, ext_id) for ext_id in external_id], sources=[self._view_id]
            ).nodes
            return self._safe_convert(nodes)

    @overload
    def delete(self, external_id: str) -> NodeId: ...

    @overload
    def delete(self, external_id: SequenceNotStr[str]) -> list[NodeId]: ...

    def delete(self, external_id: str | SequenceNotStr[str]) -> NodeId | list[NodeId]:
        """Delete a view source or a list of view sources by their external IDs."""
        if isinstance(external_id, str):
            return self._instance_api.delete(NodeId(COGNITE_MIGRATION_SPACE, external_id)).nodes[0]
        else:
            return self._instance_api.delete([NodeId(COGNITE_MIGRATION_SPACE, ext_id) for ext_id in external_id]).nodes

    def list(
        self, resource_type: str | None = None, limit: int | None = DEFAULT_LIMIT_READ
    ) -> NodeList[ResourceViewMapping]:
        """List view sources optionally filtered by resource type"""
        is_selected: filters.Filter | None = None
        if resource_type:
            is_selected = filters.Equals(self._view_id.as_property_ref("resourceType"), resource_type)

        nodes = self._instance_api.list(
            instance_type="node", filter=is_selected, limit=limit, space=COGNITE_MIGRATION_SPACE, sources=self._view_id
        )
        return self._safe_convert(nodes)

    @classmethod
    def _safe_convert(cls, nodes: NodeList[Node]) -> NodeList[ResourceViewMapping]:
        results = NodeList[ResourceViewMapping]([])
        for node in nodes:
            try:
                loaded = ResourceViewMapping._load(node.dump())
            except ValueError as e:
                warnings.warn(
                    HighSeverityWarning(
                        f"Node {node.as_id()!r} is in an invalid format. Skipping it. Error: {e!s}",
                    )
                )
                continue
            results.append(loaded)
        return results


class CreatedSourceSystemAPI:
    def __init__(self, instance_api: ExtendedInstancesAPI) -> None:
        self._instance_api = instance_api
        self._RETRIEVE_LIMIT = 1000
        self._view_id = CreatedSourceSystem.get_source()

    def retrieve(self, source: SequenceNotStr[str]) -> NodeList[CreatedSourceSystem]:
        """Retrieve one or more view sources by their external IDs."""
        results: NodeList[CreatedSourceSystem] = NodeList[CreatedSourceSystem]([])
        # MyPy does not understand that SequenceNotStr is a sequence.
        for chunk in chunker_sequence(source, self._RETRIEVE_LIMIT):  # type: ignore[type-var]
            retrieve_query = query.Query(
                with_={
                    "sourceSystem": query.NodeResultSetExpression(
                        filter=filters.And(filters.HasData(views=[self._view_id]), self._create_dms_filter(chunk)),
                        limit=len(chunk),
                    ),
                },
                select={"sourceSystem": query.Select([query.SourceSelector(self._view_id, ["*"])])},
            )
            chunk_response = self._instance_api.query(retrieve_query)
            results.extend([CreatedSourceSystem._load(item.dump()) for item in chunk_response.get("sourceSystem", [])])
        return results

    def _create_dms_filter(self, source: SequenceNotStr[str]) -> filters.Filter:
        """Create a filter that matches all CreatedSourceSystem with given source in the list."""
        if not source:
            raise ValueError("Cannot create a filter from an empty source list.")
        return filters.In(self._view_id.as_property_ref("source"), list(source))

    def list(self, limit: int = -1) -> NodeList[CreatedSourceSystem]:
        """Lists all created source systems."""
        return self._instance_api.list(instance_type=CreatedSourceSystem, limit=limit)


_T = TypeVar("_T", bound=int | str)


class SpaceSourceAPI:
    def __init__(self, instance_api: ExtendedInstancesAPI) -> None:
        self._instance_api = instance_api
        self._RETRIEVE_LIMIT = 1000
        self._view_id = SpaceSource.get_source()
        # Cache for space sources by data set ID and external ID
        self._cache_by_id: dict[int, SpaceSource] = {}
        self._cache_by_external_id: dict[str, SpaceSource] = {}

    @overload
    def retrieve(self, data_set_id: int) -> SpaceSource | None: ...

    @overload
    def retrieve(self, data_set_id: Sequence[int]) -> NodeList[SpaceSource]: ...

    @overload
    def retrieve(self, *, data_set_external_id: str) -> SpaceSource | None: ...

    @overload
    def retrieve(self, *, data_set_external_id: SequenceNotStr[str]) -> NodeList[SpaceSource]: ...

    def retrieve(
        self,
        data_set_id: int | Sequence[int] | None = None,
        data_set_external_id: str | SequenceNotStr[str] | None = None,
    ) -> SpaceSource | NodeList[SpaceSource] | None:
        """Retrieve a space source by data set ID or external ID.

        This method uses caching to avoid redundant API calls. If a space source
        has been retrieved before, it will be returned from cache.
        """
        if data_set_id is not None and data_set_external_id is None:
            return self._retrieve_with_cache(
                value=data_set_id,
                property_name="dataSetId",
                cache=self._cache_by_id,
                is_single=isinstance(data_set_id, int),
            )
        elif data_set_external_id is not None:
            return self._retrieve_with_cache(
                # MyPy is confused by SequenceNotStr
                value=data_set_external_id,  # type: ignore[arg-type]
                property_name="dataSetExternalId",
                cache=self._cache_by_external_id,
                is_single=isinstance(data_set_external_id, str),
            )
        else:
            raise ValueError("Either data_set_id or data_set_external_id must be provided, but not both.")

    def _retrieve_with_cache(
        self,
        value: _T | Sequence[_T],
        property_name: str,
        cache: dict[_T, SpaceSource],
        is_single: bool,
    ) -> SpaceSource | NodeList[SpaceSource] | None:
        """Retrieve space sources with caching support."""
        # MyPy does not understand that if is_single is True, value is _T, else Sequence[_T].
        values: list[_T] = [value] if is_single else list(value)  # type: ignore[arg-type, list-item]

        # Check cache for all requested values
        cached_results: list[SpaceSource] = []
        missing_values: list[_T] = []
        for val in values:
            if val in cache:
                cached_results.append(cache[val])
            else:
                missing_values.append(val)

        # Fetch missing values from API
        if missing_values:
            fetched = self._retrieve_by_property(property_name=property_name, values=missing_values)
            for item in fetched:
                self._add_to_cache(item)
                cached_results.append(item)

        if is_single:
            return cached_results[0] if cached_results else None
        return NodeList[SpaceSource](cached_results)

    def _add_to_cache(self, space_source: SpaceSource) -> None:
        """Add a space source to both caches."""
        if space_source.data_set_id:
            self._cache_by_id[space_source.data_set_id] = space_source
        if space_source.data_set_external_id is not None:
            self._cache_by_external_id[space_source.data_set_external_id] = space_source

    def _retrieve_by_property(
        self,
        property_name: str,
        values: Sequence[_T],
    ) -> NodeList[SpaceSource]:
        """Retrieve space sources by filtering on a specific property."""
        results: NodeList[SpaceSource] = NodeList[SpaceSource]([])
        for chunk in chunker_sequence(values, self._RETRIEVE_LIMIT):
            retrieve_query = query.Query(
                with_={
                    "spaceSource": query.NodeResultSetExpression(
                        filter=filters.And(
                            filters.HasData(views=[self._view_id]),
                            filters.In(self._view_id.as_property_ref(property_name), chunk),
                        ),
                        limit=len(chunk),
                    ),
                },
                select={"spaceSource": query.Select([query.SourceSelector(self._view_id, ["*"])])},
            )
            chunk_response = self._instance_api.query(retrieve_query)
            items = chunk_response.get("spaceSource", [])
            results.extend([SpaceSource._load(item.dump()) for item in items])

        return results

    def list(self, limit: int = -1) -> NodeList[SpaceSource]:
        """Lists all space sources and populates the cache.

        Note: This method always fetches from the API and does not use the cache,
        but it will populate the cache with all retrieved space sources.
        """
        results = self._instance_api.list(instance_type=SpaceSource, limit=limit)
        # Populate cache with all retrieved space sources
        for space_source in results:
            self._add_to_cache(space_source)
        return results


class MigrationAPI:
    def __init__(self, instance_api: ExtendedInstancesAPI) -> None:
        self.instance_source = InstanceSourceAPI(instance_api)
        self.resource_view_mapping = ResourceViewMappingAPI(instance_api)
        self.created_source_system = CreatedSourceSystemAPI(instance_api)
        self.space_source = SpaceSourceAPI(instance_api)



from collections.abc import Iterator
from pathlib import Path

import pytest
from cognite.client.data_classes import AssetList, AssetWrite, AssetWriteList
from cognite.client.data_classes.data_modeling import NodeId, Space, ViewId
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteAsset
from cognite.client.exceptions import CogniteAPIError

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands._migrate.command import MigrationCommand
from cognite_toolkit._cdf_tk.commands._migrate.data_mapper import AssetCentricMapper
from cognite_toolkit._cdf_tk.commands._migrate.default_mappings import (
    ASSET_ID,
    EVENT_ID,
    FILE_METADATA_ID,
    TIME_SERIES_ID,
)
from cognite_toolkit._cdf_tk.commands._migrate.migration_io import AssetCentricMigrationIO
from cognite_toolkit._cdf_tk.commands._migrate.selectors import MigrateDataSetSelector, MigrationCSVFileSelector
from tests.test_integration.conftest import HierarchyMinimal
from tests.test_integration.constants import RUN_UNIQUE_ID


@pytest.fixture()
def three_assets(toolkit_client: ToolkitClient, toolkit_space: Space) -> Iterator[AssetList]:
    client = toolkit_client
    space = toolkit_space.space
    assets = AssetWriteList([])
    for i in range(3):
        asset = AssetWrite(
            external_id=f"toolkit_asset_test_migration_{i}_{RUN_UNIQUE_ID}",
            name=f"toolkit_asset_test_migration_{i}_{RUN_UNIQUE_ID}",
            parent_external_id=f"toolkit_asset_test_migration_{0}_{RUN_UNIQUE_ID}" if i > 0 else None,
        )
        assets.append(asset)
    output = client.assets.retrieve_multiple(external_ids=assets.as_external_ids(), ignore_unknown_ids=True)
    if output:
        try:
            client.assets.delete(external_id=output.as_external_ids(), ignore_unknown_ids=True)
        except CogniteAPIError:
            client.data_modeling.instances.delete([NodeId(space, ts.external_id) for ts in output])
    created = client.assets.create(assets)

    yield created

    # Cleanup after test
    deleted = client.data_modeling.instances.delete([NodeId(space, ts.external_id) for ts in created])
    if deleted.nodes:
        return
    client.assets.delete(external_id=created.as_external_ids())


class TestMigrateAssetsCommand:
    def test_migrate_assets(
        self,
        toolkit_client: ToolkitClient,
        three_assets: AssetList,
        toolkit_space: Space,
        tmp_path: Path,
    ) -> None:
        client = toolkit_client
        space = toolkit_space.space

        input_file = tmp_path / "timeseries_migration.csv"
        with input_file.open("w", encoding="utf-8") as f:
            f.write(
                "id,dataSetId,space,externalId\n"
                + "\n".join(
                    f"{a.id},{a.data_set_id if a.data_set_id else ''},{space},{a.external_id}" for a in three_assets
                )
                + "\n"
            )

        cmd = MigrationCommand(skip_tracking=True, silent=True)
        cmd.migrate(
            selected=MigrationCSVFileSelector(datafile=input_file, kind="Assets"),
            data=AssetCentricMigrationIO(client),
            mapper=AssetCentricMapper(client),
            log_dir=tmp_path / "logs",
            dry_run=False,
            verbose=False,
        )
        node_ids = [NodeId(space, a.external_id) for a in three_assets]
        migrated_assets = client.data_modeling.instances.retrieve_nodes(node_ids, CogniteAsset)
        assert len(migrated_assets) == len(three_assets), "Not all assets were migrated successfully."

    def test_migrate_assets_by_dataset_dry_run(
        self, toolkit_client: ToolkitClient, migration_hierarchy_minimal: HierarchyMinimal, tmp_path: Path
    ) -> None:
        client = toolkit_client
        hierarchy = migration_hierarchy_minimal
        cmd = MigrationCommand(skip_tracking=True, silent=True)
        progress = cmd.migrate(
            selected=MigrateDataSetSelector(
                kind="Assets",
                data_set_external_id=hierarchy.dataset.external_id,
                ingestion_mapping=ASSET_ID,
                preferred_consumer_view=ViewId("cdf_cdm", "CogniteAsset", "v1"),
            ),
            data=AssetCentricMigrationIO(client),
            mapper=AssetCentricMapper(client),
            log_dir=tmp_path,
            dry_run=True,
        )
        results = progress.aggregate()
        expected_results = {(step, "success"): 2 for step in cmd.Steps.list()}
        assert results == expected_results


class TestMigrateEventsCommand:
    def test_migrate_events_by_dataset_dry_run(
        self, toolkit_client: ToolkitClient, migration_hierarchy_minimal: HierarchyMinimal, tmp_path: Path
    ) -> None:
        client = toolkit_client
        hierarchy = migration_hierarchy_minimal
        cmd = MigrationCommand(skip_tracking=True, silent=True)
        progress = cmd.migrate(
            selected=MigrateDataSetSelector(
                kind="Events",
                data_set_external_id=hierarchy.dataset.external_id,
                ingestion_mapping=EVENT_ID,
                preferred_consumer_view=ViewId("cdf_cdm", "CogniteActivity", "v1"),
            ),
            data=AssetCentricMigrationIO(client),
            mapper=AssetCentricMapper(client),
            log_dir=tmp_path,
            dry_run=True,
        )
        results = progress.aggregate()
        expected_results = {(step, "success"): 1 for step in cmd.Steps.list()}
        assert results == expected_results


class TestMigrateTimeSeriesCommand:
    def test_migrate_time_series_by_dataset_dry_run(
        self, toolkit_client: ToolkitClient, migration_hierarchy_minimal: HierarchyMinimal, tmp_path: Path
    ) -> None:
        client = toolkit_client
        hierarchy = migration_hierarchy_minimal
        cmd = MigrationCommand(skip_tracking=True, silent=True)
        progress = cmd.migrate(
            selected=MigrateDataSetSelector(
                kind="TimeSeries",
                data_set_external_id=hierarchy.dataset.external_id,
                ingestion_mapping=TIME_SERIES_ID,
                preferred_consumer_view=ViewId("cdf_cdm", "CogniteTimeSeries", "v1"),
            ),
            data=AssetCentricMigrationIO(client, skip_linking=True),
            mapper=AssetCentricMapper(client),
            log_dir=tmp_path,
            dry_run=True,
        )
        results = progress.aggregate()
        expected_results = {(step, "success"): 1 for step in cmd.Steps.list()}
        assert results == expected_results


class TestMigrateFileMetadataCommand:
    def test_migrate_file_metadata_by_dataset_dry_run(
        self, toolkit_client: ToolkitClient, migration_hierarchy_minimal: HierarchyMinimal, tmp_path: Path
    ) -> None:
        client = toolkit_client
        hierarchy = migration_hierarchy_minimal
        cmd = MigrationCommand(skip_tracking=True, silent=True)
        progress = cmd.migrate(
            selected=MigrateDataSetSelector(
                kind="FileMetadata",
                data_set_external_id=hierarchy.dataset.external_id,
                ingestion_mapping=FILE_METADATA_ID,
                preferred_consumer_view=ViewId("cdf_cdm", "CogniteFile", "v1"),
            ),
            data=AssetCentricMigrationIO(client, skip_linking=False),
            mapper=AssetCentricMapper(client),
            log_dir=tmp_path,
            dry_run=True,
        )
        results = progress.aggregate()
        expected_results = {(step, "success"): 1 for step in cmd.Steps.list()}
        assert results == expected_results



import ctypes
import json
from abc import ABC, abstractmethod
from collections.abc import Mapping
from datetime import datetime
from typing import ClassVar, Literal, overload

from cognite.client.data_classes import Label, LabelDefinition
from cognite.client.data_classes.data_modeling import ContainerId
from cognite.client.data_classes.data_modeling.data_types import (
    DirectRelationReference,
    Enum,
    ListablePropertyType,
    PropertyType,
)
from cognite.client.data_classes.data_modeling.instances import PropertyValueWrite
from cognite.client.utils import ms_to_datetime
from dateutil import parser

from cognite_toolkit._cdf_tk.constants import CDF_UNIT_SPACE
from cognite_toolkit._cdf_tk.exceptions import ToolkitNotSupported
from cognite_toolkit._cdf_tk.utils._auxiliary import get_concrete_subclasses
from cognite_toolkit._cdf_tk.utils.useful_types import AssetCentricType, DataType, JsonVal, PythonTypes

from .collection import humanize_collection

INT32_MIN = -2_147_483_648
INT32_MAX = 2_147_483_647
INT64_MIN = -9_223_372_036_854_775_808
INT64_MAX = 9_223_372_036_854_775_807


def asset_centric_convert_to_primary_property(
    value: str | int | float | bool | dict | list | None,
    type_: PropertyType,
    nullable: bool,
    destination_container_property: tuple[ContainerId, str],
    source_property: tuple[AssetCentricType, str],
    direct_relation_lookup: Mapping[str | int, DirectRelationReference] | None = None,
) -> PropertyValueWrite:
    if (source_property, destination_container_property) in SPECIAL_CONVERTER_BY_SOURCE_DESTINATION:
        converter_cls = SPECIAL_CONVERTER_BY_SOURCE_DESTINATION[(source_property, destination_container_property)]
        if issubclass(converter_cls, _SourceConverter):
            if direct_relation_lookup is None:
                raise TypeError("Direct relation lookup must be provided for DirectRelationReference conversion")
            return converter_cls(direct_relation_lookup=direct_relation_lookup).convert(value)
        else:
            return converter_cls().convert(value)
    else:
        # Fallback to the standard conversion
        return convert_to_primary_property(value, type_, nullable, direct_relation_lookup=direct_relation_lookup)


def convert_to_primary_property(
    value: str | int | float | bool | dict | list | None,
    type_: PropertyType,
    nullable: bool,
    direct_relation_lookup: Mapping[str | int, DirectRelationReference] | None = None,
) -> PropertyValueWrite:
    """Convert a string value to the appropriate type based on the provided property type.

    Args:
        value (str | int | float | bool): The value to convert.
        type_ (PropertyType): The type of the property to convert to.
        nullable (bool): Whether the property can be null.
        direct_relation_lookup (Mapping[str | int, DirectRelationReference] | None): A mapping from external/internal
            IDs to DirectRelationReference objects,

    Returns:
        PropertyValueWrite: The converted value as a PropertyValue, or None if is_nullable is True and the value is empty.
    """
    dtype = type_._type
    if dtype in CONVERTER_BY_DTYPE:
        converter_cls = CONVERTER_BY_DTYPE[dtype]
    else:
        raise TypeError(f"Unsupported property type {dtype}")
    if issubclass(converter_cls, _EnumConverter) and isinstance(type_, Enum):
        converter: _ValueConverter = _EnumConverter(type_=type_, nullable=nullable)
    elif issubclass(converter_cls, _EnumConverter):
        raise TypeError(f"Unsupported property type {type_} for enum conversion")
    elif issubclass(converter_cls, _DirectRelationshipConverter):
        if direct_relation_lookup is None:
            raise TypeError("Direct relation lookup must be provided for DirectRelationReference conversion")
        converter = _DirectRelationshipConverter(direct_relation_lookup=direct_relation_lookup)
    else:
        converter = converter_cls(nullable)

    if isinstance(type_, ListablePropertyType) and type_.is_list:
        values = _as_list(value)
        output: list[PropertyValueWrite] = []
        for item in values:
            converted = converter.convert(item)  # type: ignore[arg-type]
            if converted is not None:
                output.append(converted)
        # MyPy gets confused by the SequenceNotStr used in the PropertyValueWrite
        return output  # type: ignore[return-value]
    else:
        return converter.convert(value)


def convert_str_to_data_type(
    value: str | None,
    type_: DataType,
    nullable: bool = True,
    is_array: bool = False,
) -> PythonTypes | None:
    """Convert a string value to the appropriate data type based on the provided type.

    Args:
        value: The value to convert, which can be a string or None.
        type_: The target data type.
        nullable: Whether data type can be null.
        is_array: Whether the data type is an array.

    Returns:
        The converted value as a string, int, float, bool, datetime, date, dict, list, or None.

    """
    if type_ not in DATATYPE_CONVERTER_BY_DATA_TYPE:
        raise ValueError(
            f"Unsupported data type {type_}. Available types: {humanize_collection(DATATYPE_CONVERTER_BY_DATA_TYPE.keys())}."
        )
    converter_cls = DATATYPE_CONVERTER_BY_DATA_TYPE[type_]
    converter = converter_cls(nullable)
    if is_array:
        values = _as_list(value)
        output: list[PythonTypes] = []
        for item in values:
            converted = converter.convert(item)  # type: ignore[arg-type]
            if converted is not None:
                output.append(converted)  # type: ignore[arg-type]
        # MyPy gets confused by the SequenceNotStr used in the PropertyValueWrite
        return output
    else:
        return converter.convert(value)  # type: ignore[return-value]


@overload
def infer_data_type_from_value(value: str, dtype: Literal["Json"]) -> tuple[DataType, JsonVal]: ...


@overload
def infer_data_type_from_value(value: str, dtype: Literal["Python"]) -> tuple[DataType, PythonTypes]: ...


def infer_data_type_from_value(value: str, dtype: Literal["Json", "Python"]) -> tuple[DataType, JsonVal | PythonTypes]:
    """Infer the data type from a given value.

    Args:
        value: The value to infer the data type from, which can be a string.
        dtype: The data type to infer. Can be "JsonVal" or "Python". Python type will infer datetime and date types
            as well.

    Returns:
        A tuple containing the inferred data type and the converted value.

    """
    converter_classes = (
        _Int64Converter,
        _Float64Converter,
        _TimestampConverter,
        _BooleanConverter,
        _JsonConverter,
        _TextConverter,
    )
    converters_to_use = (
        tuple(converter_cls for converter_cls in converter_classes if converter_cls is not _TimestampConverter)
        if dtype == "Json"
        else converter_classes
    )
    for converter_cls in converters_to_use:
        # MyPy thinks that converter_cls can be abstract, but it is not
        converter = converter_cls(nullable=False)
        try:
            converted_value = converter.convert(value)
        except ValueError:
            continue

        if (
            converter_cls is _TimestampConverter
            and isinstance(converted_value, datetime)
            and _is_midnight_and_naive(converted_value)
        ):
            # If the converted value is a datetime with no time component, return it as a date
            return _DateConverter.schema_type, converted_value.date()
        else:
            return converter_cls.schema_type, converted_value  # type: ignore[return-value]

    raise ValueError(
        f"Failed to infer data type from value: {value!r}. Supported types are: "
        f"{humanize_collection(DATATYPE_CONVERTER_BY_DATA_TYPE.keys())}."
    )


def _is_midnight_and_naive(dt: datetime) -> bool:
    """Checks if a datetime object is at midnight and is naive (has no timezone)."""
    return not (dt.hour or dt.minute or dt.second or dt.microsecond or dt.tzinfo)


def _as_list(value: str | int | float | bool | dict[str, object] | list[object] | None) -> list[object]:
    """Convert a value to a list, ensuring that it is iterable."""
    if value is None:
        return []
    elif isinstance(value, list):
        return value
    elif isinstance(value, str) and value.strip() == "":
        return []
    elif isinstance(value, str):
        try:
            data = json.loads(value)
            return data if isinstance(data, list) else [data]
        except json.JSONDecodeError:
            return [value]
    elif isinstance(value, int | float | bool | dict):
        return [value]
    else:
        raise TypeError(f"Cannot convert {value} of type {type(value)} to a list.")


class _Converter(ABC):
    @abstractmethod
    def convert(self, value: str | int | float | bool | dict | list | None) -> PropertyValueWrite:
        """Convert a value to the appropriate type."""
        raise NotImplementedError("This method should be implemented by subclasses.")


class _ValueConverter(_Converter, ABC):
    type_str: ClassVar[str]
    schema_type: ClassVar[DataType | None] = None
    _handles_list: ClassVar[bool] = False

    def __init__(self, nullable: bool) -> None:
        self.nullable = nullable

    def convert(self, value: str | int | float | bool | dict | list | None) -> PropertyValueWrite:
        if value is None and self.nullable is False:
            raise ValueError("Cannot convert None to a non-nullable property.")
        elif value is None:
            return None
        elif isinstance(value, list) and not self._handles_list:
            raise ValueError(f"Expected a single value for {self.type_str}, but got a list.")
        # If the value is a list, we handle it in the subclass if it supports lists.
        return self._convert(value)  # type: ignore[arg-type]

    @abstractmethod
    def _convert(self, value: str | int | float | bool | dict) -> PropertyValueWrite:
        """Convert the value to the appropriate type."""
        raise NotImplementedError("This method should be implemented by subclasses.")


class _SpecialCaseConverter(_Converter, ABC):
    """Abstract base class for converters handling special cases."""

    source_property: ClassVar[tuple[AssetCentricType, str]]
    destination_container_property: ClassVar[tuple[ContainerId, str]]


class _TimeSeriesTypeConverter(_SpecialCaseConverter):
    source_property = ("timeseries", "isString")
    destination_container_property = (ContainerId("cdf_cdm", "CogniteTimeSeries"), "type")

    def convert(self, value: str | int | float | bool | dict | list | None) -> str:
        if isinstance(value, bool):
            return "string" if value else "numeric"
        raise ValueError(f"Cannot convert {value} to TimeSeries type. Expected a boolean value.")


class _TimeSeriesUnitConverter(_SpecialCaseConverter):
    source_property = ("timeseries", "unitExternalId")
    destination_container_property = (ContainerId("cdf_cdm", "CogniteTimeSeries"), "unit")

    def convert(self, value: str | int | float | bool | dict | list | None) -> dict[str, str] | None:
        if value is None:
            return None
        if isinstance(value, str):
            return {"space": CDF_UNIT_SPACE, "externalId": value}
        raise ValueError(f"Cannot convert {value!r} to TimeSeries unit. Expected a string representing the externalId.")


class _LabelConverter(_SpecialCaseConverter, ABC):
    destination_container_property = (ContainerId("cdf_cdm", "CogniteDescribable"), "tags")

    def convert(self, value: str | int | float | bool | dict | list | None) -> list[str]:
        if not isinstance(value, list):
            raise ValueError(
                f"Cannot convert {value} to labels. Expected a list of Labels, objects, or LabelDefinitions."
            )
        tags: list[str] = []
        for item in value:
            match item:
                case str() as tag:
                    tags.append(tag)
                # Matches a dict with the key "externalId" whose value is a string
                case {"externalId": str() as tag}:
                    tags.append(tag)
                # Matches Label or LabelDefinition objects with a truthy external_id
                case Label(external_id=tag) | LabelDefinition(external_id=tag) if tag:
                    tags.append(tag)
                case _:
                    raise ValueError(
                        f"Invalid label item: {item}. Expected a string, dict with 'externalId', or Label/LabelDefinition."
                    )
        return tags


class _AssetLabelConverter(_LabelConverter):
    source_property = ("asset", "labels")


class _FileLabelConverter(_LabelConverter):
    source_property = ("file", "labels")


class _SourceConverter(_SpecialCaseConverter, ABC):
    destination_container_property = (ContainerId("cdf_cdm", "CogniteSourceable"), "source")

    def __init__(self, direct_relation_lookup: Mapping[str | int, DirectRelationReference]) -> None:
        self.direct_relation_lookup = direct_relation_lookup

    def convert(self, value: str | int | float | bool | dict | list | None) -> DirectRelationReference:
        if isinstance(value, str | int):
            if value in self.direct_relation_lookup:
                return self.direct_relation_lookup[value]
            elif isinstance(value, str) and value.casefold() in self.direct_relation_lookup:
                # The aggregate endpoint used to create source system lowercases all sources, thus
                # we do a case-insensitive lookup as a fallback.
                return self.direct_relation_lookup[value.casefold()]
        raise ValueError(
            f"Cannot convert {value!r} to DirectRelationReference. Invalid data type or missing in lookup."
        )


class _AssetSourceConverter(_SourceConverter):
    source_property = ("asset", "source")


class _FileSourceConverter(_SourceConverter):
    source_property = ("file", "source")


class _EventSourceConverter(_SourceConverter):
    source_property = ("event", "source")


class _TextConverter(_ValueConverter):
    type_str = "text"
    schema_type = "string"

    def _convert(self, value: str | int | float | bool | dict) -> PropertyValueWrite:
        return str(value)


class _BooleanConverter(_ValueConverter):
    type_str = "boolean"
    schema_type = "boolean"

    def _convert(self, value: str | int | float | bool | dict) -> PropertyValueWrite:
        if isinstance(value, bool | int | float):
            return bool(value)
        elif isinstance(value, str):
            if value.lower() in ("true", "1"):
                return True
            elif value.lower() in ("false", "0"):
                return False
        raise ValueError(f"Cannot convert {value} to boolean.")


class _Int32Converter(_ValueConverter):
    type_str = "int32"

    def _convert(self, value: str | int | float | bool | dict) -> PropertyValueWrite:
        if isinstance(value, int):
            output = value
        elif isinstance(value, str):
            try:
                output = int(value)
            except ValueError:
                raise ValueError(f"Cannot convert {value} to int32.")
        else:
            raise ValueError(f"Cannot convert {value} to int32.")
        if output < INT32_MIN or output > INT32_MAX:
            raise ValueError(f"Value {value} is out of range for int32.")
        return output


class _Int64Converter(_ValueConverter):
    type_str = "int64"
    schema_type = "integer"

    def _convert(self, value: str | int | float | bool | dict) -> PropertyValueWrite:
        if isinstance(value, int):
            output = value
        elif isinstance(value, str):
            try:
                output = int(value)
            except ValueError:
                raise ValueError(f"Cannot convert {value} to int64.")
        else:
            raise ValueError(f"Cannot convert {value} to int64.")
        if not (INT64_MIN <= output <= INT64_MAX):
            raise ValueError(f"Value {value} is out of range for int64.")
        return output


class _Float32Converter(_ValueConverter):
    type_str = "float32"

    def _convert(self, value: str | int | float | bool | dict) -> PropertyValueWrite:
        if isinstance(value, float | int):
            output = float(value)
        elif isinstance(value, str):
            try:
                output = float(value)
            except ValueError:
                raise ValueError(f"Cannot convert {value} to float32.")
        else:
            raise ValueError(f"Cannot convert {value} to float32.")
        try:
            output = ctypes.c_float(output).value
        except OverflowError:
            raise ValueError(f"Value {value} is out of range for float32.")
        if output == float("inf") or output == float("-inf"):
            raise ValueError(f"Value {value} is out of range for float32.")
        return output


class _Float64Converter(_ValueConverter):
    type_str = "float64"
    schema_type = "float"

    def _convert(self, value: str | int | float | bool | dict) -> PropertyValueWrite:
        if isinstance(value, float | int):
            output = float(value)
        elif isinstance(value, str):
            try:
                output = float(value)
            except ValueError:
                raise ValueError(f"Cannot convert {value} to float64.")
        else:
            raise ValueError(f"Cannot convert {value} to float64.")
        if output == float("inf") or output == float("-inf"):
            raise ValueError(f"Value {value} is out of range for float64.")
        return output


class _JsonConverter(_ValueConverter):
    type_str = "json"
    schema_type = "json"
    _handles_list = True

    def _convert(self, value: str | int | float | bool | dict[str, object] | list) -> PropertyValueWrite:
        if isinstance(value, bool | int | float):
            return value
        elif isinstance(value, dict):
            if non_string_keys := [k for k in value if not isinstance(k, str)]:
                raise ValueError(
                    f"JSON keys must be strings. Found non-string keys: {humanize_collection(non_string_keys)}"
                )
            return value
        elif isinstance(value, list):
            if not all(isinstance(item, str | int | float | bool | dict | list) for item in value):
                raise ValueError("All items in the list must be of type str, int, float, bool, dict, or list.")
            return value
        elif isinstance(value, str):
            try:
                return json.loads(value)
            except json.JSONDecodeError as e:
                raise ValueError(f"Cannot convert {value} to JSON: {e}") from e
        raise ValueError(f"Cannot convert {value} to JSON.")


class _TimestampConverter(_ValueConverter):
    type_str = "timestamp"
    schema_type = "timestamp"

    def _convert(self, value: str | int | float | bool | dict) -> datetime:
        if isinstance(value, int | float):
            try:
                return ms_to_datetime(value)
            except (ValueError, OSError) as e:
                raise ValueError(f"Cannot convert numeric value {value} to timestamp: {e}") from e
        elif isinstance(value, str):
            try:
                return parser.isoparse(value)
            except ValueError as e:
                raise ValueError(f"Cannot convert {value} to timestamp: {e}") from e
        raise ValueError(f"Cannot convert {value} to timestamp.")


class _EpochConverter(_ValueConverter):
    type_str = "epoch"
    schema_type = "epoch"

    def _convert(self, value: str | int | float | bool | dict) -> int:
        if isinstance(value, int | float):
            return int(value)
        elif isinstance(value, str):
            try:
                dt = parser.isoparse(value)
                epoch = int(dt.timestamp() * 1000)
                return epoch
            except ValueError as e:
                raise ValueError(f"Cannot convert {value} to epoch timestamp: {e}") from e
        raise ValueError(f"Cannot convert {value} to epoch timestamp.")


class _DateConverter(_ValueConverter):
    type_str = "date"
    schema_type = "date"

    def _convert(self, value: str | int | float | bool | dict) -> PropertyValueWrite:
        if isinstance(value, str):
            try:
                return parser.parse(value).date()
            except ValueError as e:
                raise ValueError(f"Cannot convert {value} to date: {e}") from e
        raise ValueError(f"Cannot convert {value} to date.")


class _EnumConverter(_ValueConverter):
    type_str = "enum"

    def __init__(self, type_: Enum, nullable: bool) -> None:
        super().__init__(nullable)
        self.available_types = {enum_value.casefold(): enum_value for enum_value in type_.values.keys()}

    def _convert(self, value: str | int | float | bool | dict) -> PropertyValueWrite:
        value = str(value).casefold()
        if value in self.available_types:
            return self.available_types[value]
        raise ValueError(
            f"Value {value!r} is not a valid enum value. Available values: {humanize_collection(self.available_types.values())}"
        )


class _DirectRelationshipConverter(_ValueConverter):
    type_str = "direct"

    def __init__(self, direct_relation_lookup: Mapping[str | int, DirectRelationReference]) -> None:
        super().__init__(nullable=True)
        self.direct_relation_lookup = direct_relation_lookup

    def _convert(self, value: str | int | float | bool | dict) -> PropertyValueWrite:
        if isinstance(value, str | int) and value in self.direct_relation_lookup:
            return self.direct_relation_lookup[value]
        raise ValueError(
            f"Cannot convert {value!r} to DirectRelationReference. Invalid data type or missing in lookup."
        )


class _TimeSeriesReferenceConverter(_ValueConverter):
    type_str = "timeseries"

    def _convert(self, value: str | int | float | bool | dict) -> PropertyValueWrite:
        raise ToolkitNotSupported("Timeseries reference conversion is not supported.")


class _FileReferenceConverter(_ValueConverter):
    type_str = "file"

    def _convert(self, value: str | int | float | bool | dict) -> PropertyValueWrite:
        raise ToolkitNotSupported("File reference conversion is not supported.")


class _SequenceReferenceConverter(_ValueConverter):
    type_str = "sequence"

    def _convert(self, value: str | int | float | bool | dict) -> PropertyValueWrite:
        raise ToolkitNotSupported("Sequence reference conversion is not supported.")


CONVERTER_BY_DTYPE: Mapping[str, type[_ValueConverter]] = {
    cls_.type_str: cls_  # type: ignore[type-abstract]
    for cls_ in _ValueConverter.__subclasses__()
}
SPECIAL_CONVERTER_BY_SOURCE_DESTINATION: Mapping[
    tuple[tuple[AssetCentricType, str], tuple[ContainerId, str]],
    type[_SpecialCaseConverter],
] = {
    (subclass.source_property, subclass.destination_container_property): subclass
    for subclass in get_concrete_subclasses(_SpecialCaseConverter)  # type: ignore[type-abstract]
}
DATATYPE_CONVERTER_BY_DATA_TYPE: Mapping[DataType, type[_ValueConverter]] = {
    cls_.schema_type: cls_
    for cls_ in get_concrete_subclasses(_ValueConverter)  # type: ignore[type-abstract]
    if cls_.schema_type is not None
}



from pathlib import Path

from cognite.client.data_classes import AggregateResultItem, Asset, AssetList, TransformationPreviewResult

from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.commands import DownloadCommand
from cognite_toolkit._cdf_tk.storageio import AssetIO
from cognite_toolkit._cdf_tk.storageio.selectors import DataSetSelector
from cognite_toolkit._cdf_tk.utils.fileio import CSVReader


class TestDownloadCommand:
    def test_download_data_in_table_format(self, tmp_path: Path) -> None:
        cmd = DownloadCommand(silent=True, skip_tracking=True)
        with monkeypatch_toolkit_client() as client:
            dataset = "my/:_data_set"
            client.assets.aggregate_count.return_value = 1
            client.assets.return_value = [
                AssetList(
                    [
                        Asset(
                            id=123,
                            name="asset_123",
                            metadata={"key": "value"},
                            data_set_id=42,
                            aggregates=AggregateResultItem(1, depth=0),
                        )
                    ]
                )
            ]
            client.lookup.data_sets.id.return_value = 42
            client.lookup.data_sets.external_id.return_value = dataset
            client.transformations.preview.return_value = TransformationPreviewResult(
                None, results=[{"key": "key", "key_count": 1}]
            )

            cmd.download(
                selectors=[DataSetSelector(kind="Assets", data_set_external_id=dataset)],
                io=AssetIO(client=client),
                output_dir=tmp_path,
                verbose=True,
                file_format=".csv",
                compression="none",
                limit=-1,
            )

            csv_files = list(tmp_path.rglob("*.csv"))
            assert len(csv_files) == 1
            content = list(CSVReader(csv_files[0]).read_chunks_unprocessed())
            assert content == [
                {
                    "dataSetExternalId": dataset,
                    "description": "",
                    "externalId": "",
                    "geoLocation": "",
                    "labels": "",
                    "metadata.key": "value",
                    "name": "asset_123",
                    "parentExternalId": "",
                    "source": "",
                    "childCount": "1",
                    "depth": "0",
                    "path": "",
                }
            ]



from abc import ABC
from dataclasses import dataclass

# We need to import Sequence from typing and not collections.abc for
# cognite_toolkit._parameters.read_parameter_from_init_type_hints to work on Python 3.9
# is necessary to avoid Ruff changing the import
from typing import Any, Literal, Sequence  # noqa

from cognite.client import CogniteClient
from cognite.client.data_classes._base import (
    CogniteObject,
    CogniteResourceList,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)
from cognite.client.data_classes.data_modeling.ids import DataModelId, ViewId
from typing_extensions import Self


@dataclass
class LocationFilterScene(CogniteObject):
    external_id: str
    space: str

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            external_id=resource["externalId"],
            space=resource["space"],
        )


@dataclass
class LocationFilterDataModel(CogniteObject):
    external_id: str
    space: str
    version: str

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            external_id=resource["externalId"],
            space=resource["space"],
            version=resource["version"],
        )


@dataclass(frozen=True)
class LocationFilterView(ViewId):
    represents_entity: Literal["MAINTENANCE_ORDER", "OPERATION", "NOTIFICATION", "ASSET"] | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            external_id=resource["externalId"],
            space=resource["space"],
            version=resource["version"],
            represents_entity=resource.get("representsEntity"),
        )

    def dump(self, camel_case: bool = True, include_type: bool = False) -> dict[str, str]:
        if include_type:
            raise ValueError(f"{type(self).__name__} cannot be dumped with type")
        output = super().dump(camel_case, False)
        if self.represents_entity:
            output["representsEntity"] = self.represents_entity
        return output


@dataclass
class AssetCentricSubFilter(CogniteObject):
    data_set_ids: list[int] | None = None
    asset_subtree_ids: list[dict[Literal["externalId", "id"], int | str]] | None = None
    external_id_prefix: str | None = None


@dataclass
class AssetCentricFilter(CogniteObject):
    assets: AssetCentricSubFilter | None = None
    events: AssetCentricSubFilter | None = None
    files: AssetCentricSubFilter | None = None
    timeseries: AssetCentricSubFilter | None = None
    sequences: AssetCentricSubFilter | None = None
    data_set_ids: list[int] | None = None
    asset_subtree_ids: list[dict[Literal["externalId", "id"], int | str]] | None = None
    external_id_prefix: str | None = None

    @classmethod
    def _load_subfilter(cls, resource: dict[str, Any], key: str) -> AssetCentricSubFilter | None:
        if key in resource:
            return AssetCentricSubFilter.load(resource[key])
        return None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            assets=cls._load_subfilter(resource, "assets"),
            events=cls._load_subfilter(resource, "events"),
            files=cls._load_subfilter(resource, "files"),
            timeseries=cls._load_subfilter(resource, "timeseries"),
            sequences=cls._load_subfilter(resource, "sequences"),
            data_set_ids=resource.get("dataSetIds"),
            asset_subtree_ids=resource.get("assetSubtreeIds"),
            external_id_prefix=resource.get("externalIdPrefix"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if self.assets:
            output["assets"] = self.assets.dump(camel_case)
        if self.events:
            output["events"] = self.events.dump(camel_case)
        if self.files:
            output["files"] = self.files.dump(camel_case)
        if self.timeseries:
            output["timeseries"] = self.timeseries.dump(camel_case)
        if self.sequences:
            output["sequences"] = self.sequences.dump(camel_case)
        return output


class LocationFilterCore(WriteableCogniteResource["LocationFilterWrite"], ABC):
    """
    LocationFilter contains information for a single LocationFilter.

    Args:
        external_id: The external ID provided by the client. Must be unique for the resource type.
        name: The name of the location filter
        parent_id: The ID of the parent location filter
        description: The description of the location filter
        data_models: The data models in the location filter
        instance_spaces: The list of spaces that instances are in
        scene: The scene config for the location filter
        asset_centric: The filter definition for asset centric resource types
        views: The view mappings for the location filter
    """

    def __init__(
        self,
        external_id: str,
        name: str,
        parent_id: int | None = None,
        description: str | None = None,
        data_models: list[DataModelId] | None = None,
        instance_spaces: list[str] | None = None,
        scene: LocationFilterScene | None = None,
        asset_centric: AssetCentricFilter | None = None,
        views: list[LocationFilterView] | None = None,
        data_modeling_type: Literal["HYBRID", "DATA_MODELING_ONLY"] | None = None,
        _parent_external_id: str | None = None,
    ) -> None:
        self.external_id = external_id
        self.name = name
        self.parent_id = parent_id
        self.description = description
        self.data_models = data_models
        self.instance_spaces = instance_spaces
        self.scene = scene
        self.asset_centric = asset_centric
        self.views = views
        self.data_modeling_type = data_modeling_type
        self._parent_external_id = _parent_external_id  # This is an internal property used to temporarily hold the parentExternalId when lookup is deferred

    def as_write(self) -> "LocationFilterWrite":
        return LocationFilterWrite(
            external_id=self.external_id,
            name=self.name,
            parent_id=self.parent_id,
            description=self.description,
            data_models=self.data_models,
            instance_spaces=self.instance_spaces,
            scene=self.scene,
            asset_centric=self.asset_centric,
            views=self.views,
            data_modeling_type=self.data_modeling_type,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if self.data_models:
            output["dataModels" if camel_case else "data_models"] = [
                data_model.dump(camel_case, include_type=False) for data_model in self.data_models
            ]
        if self.scene:
            output["scene"] = self.scene.dump(camel_case)
        if self.asset_centric:
            output["assetCentric"] = self.asset_centric.dump(camel_case)
        if self.views:
            output["views"] = [view.dump(camel_case) for view in self.views]
        return output


class LocationFilterWrite(LocationFilterCore):
    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        scene = LocationFilterScene.load(resource.get("scene", {})) if resource.get("scene") else None
        data_models = (
            [DataModelId.load(item) for item in resource.get("dataModels", {})] if resource.get("dataModels") else None
        )
        asset_centric = (
            AssetCentricFilter.load(resource.get("assetCentric", {})) if resource.get("assetCentric") else None
        )
        views = [LocationFilterView._load(view) for view in resource["views"]] if "views" in resource else None
        return cls(
            external_id=resource["externalId"],
            name=resource["name"],
            parent_id=resource.get("parentId"),
            description=resource.get("description"),
            data_models=data_models,
            instance_spaces=resource.get("instanceSpaces"),
            scene=scene,
            asset_centric=asset_centric,
            views=views,
            data_modeling_type=resource.get("dataModelingType"),
            _parent_external_id=resource.get(
                "_parentExternalId"
            ),  # This is an internal property used to temporarily hold the parentExternalId when lookup is deferred
        )


class LocationFilter(LocationFilterCore):
    """
    LocationFilter contains information for a single LocationFilter

    Args:
        id: The ID of the location filter
        external_id: The external ID provided by the client. Must be unique for the resource type.
        name: The name of the location filter
        parent_id: The ID of the parent location filter
        description: The description of the location filter
        data_models: The data models in the location filter
        instance_spaces: The list of spaces that instances are in
        scene: The scene config for the location filter
        asset_centric: The filter definition for asset centric resource types
        views: The view mappings for the location filter
        locations: The list of child locations
    """

    def __init__(
        self,
        id: int,
        external_id: str,
        name: str,
        created_time: int,
        updated_time: int,
        parent_id: int | None = None,
        description: str | None = None,
        data_models: list[DataModelId] | None = None,
        instance_spaces: list[str] | None = None,
        scene: LocationFilterScene | None = None,
        asset_centric: AssetCentricFilter | None = None,
        views: list[LocationFilterView] | None = None,
        data_modeling_type: Literal["HYBRID", "DATA_MODELING_ONLY"] | None = None,
        locations: "LocationFilterList | None" = None,
    ) -> None:
        super().__init__(
            external_id,
            name,
            parent_id,
            description,
            data_models,
            instance_spaces,
            scene,
            asset_centric,
            views,
            data_modeling_type,
        )
        self.id = id
        self.created_time = created_time
        self.updated_time = updated_time
        self.locations = locations

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            id=resource["id"],
            external_id=resource["externalId"],
            name=resource["name"],
            parent_id=resource.get("parentId"),
            description=resource.get("description"),
            data_models=[DataModelId.load(item) for item in resource["dataModels"]]
            if "dataModels" in resource
            else None,
            instance_spaces=resource.get("instanceSpaces"),
            scene=LocationFilterScene._load(resource["scene"]) if "scene" in resource else None,
            asset_centric=AssetCentricFilter._load(resource["assetCentric"]) if "assetCentric" in resource else None,
            views=[LocationFilterView._load(view) for view in resource["views"]] if "views" in resource else None,
            created_time=resource["createdTime"],
            updated_time=resource["lastUpdatedTime"],
            data_modeling_type=resource.get("dataModelingType"),
            locations=LocationFilterList._load([item for item in resource["locations"]])
            if "locations" in resource
            else None,
        )


class LocationFilterWriteList(CogniteResourceList):
    _RESOURCE = LocationFilterWrite


class LocationFilterList(WriteableCogniteResourceList[LocationFilterWrite, LocationFilter]):
    _RESOURCE = LocationFilter

    def as_write(self) -> LocationFilterWriteList:
        return LocationFilterWriteList([LocationFilter.as_write() for LocationFilter in self])



import os
from typing import Any
from unittest import mock

import pytest

from cognite_toolkit._cdf_tk.exceptions import ToolkitMissingValueError
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables

PROJECT_AND_CLUSTER = {
    "CDF_CLUSTER": "toolkit-cluster",
    "CDF_PROJECT": "the-toolkit-project",
}


class TestEnvironmentVariables:
    @pytest.mark.parametrize(
        "args",
        [
            pytest.param({**PROJECT_AND_CLUSTER, "LOGIN_FLOW": "token", "CDF_TOKEN": "super-token"}, id="token flow"),
            pytest.param(
                {
                    **PROJECT_AND_CLUSTER,
                    "LOGIN_FLOW": "client_credentials",
                    "PROVIDER": "entra_id",
                    "IDP_TENANT_ID": "my_tenant.onmicrosoft.com",
                    "IDP_CLIENT_ID": "my-identifier",
                    "IDP_CLIENT_SECRET": "my***super***secret",
                },
                id="client-credentials entra",
            ),
            pytest.param(
                {
                    **PROJECT_AND_CLUSTER,
                    "LOGIN_FLOW": "client_credentials",
                    "PROVIDER": "other",
                    "IDP_TOKEN_URL": "https://auth.login.my_domian.io/oauth/token",
                    "IDP_AUDIENCE": "https://toolkit-cluster.fusion.cognite.com/the-toolkit-project",
                    "IDP_SCOPES": "IDENTITY,ADMIN,client:cognite-cicd@my_domain.io,user_impersonation",
                    "IDP_CLIENT_ID": "my-identifier",
                    "IDP_CLIENT_SECRET": "my***super***secret",
                },
                id="client-credentials other (auth0)",
            ),
            pytest.param(
                {
                    **PROJECT_AND_CLUSTER,
                    "LOGIN_FLOW": "interactive",
                    "PROVIDER": "entra_id",
                    "IDP_TENANT_ID": "my_tenant.onmicrosoft.com",
                    "IDP_CLIENT_ID": "my-identifier",
                },
                id="interactive entra",
            ),
            pytest.param(
                {
                    **PROJECT_AND_CLUSTER,
                    "LOGIN_FLOW": "device_code",
                    "PROVIDER": "entra_id",
                    "IDP_TENANT_ID": "my_tenant.onmicrosoft.com",
                },
                id="device enta",
            ),
            pytest.param(
                {
                    **PROJECT_AND_CLUSTER,
                    "LOGIN_FLOW": "device_code",
                    "PROVIDER": "other",
                    "IDP_CLIENT_ID": "my-identifier",
                    "IDP_DISCOVERY_URL": "https://auth.login.my_domian.io/oauth",
                },
                id="device other",
            ),
            pytest.param(
                {
                    **PROJECT_AND_CLUSTER,
                    "LOGIN_FLOW": "client_credentials",
                    "PROVIDER": "cdf",
                    "IDP_CLIENT_ID": "my-identifier",
                    "IDP_CLIENT_SECRET": "my-secret",
                },
                id="client-credentials cdf",
            ),
        ],
    )
    def test_get_valid_config(self, args: dict[str, Any]) -> None:
        env_vars = EnvironmentVariables(**args)

        missing_env_vars = env_vars.get_missing_vars()
        assert not missing_env_vars, f"Missing environment variables: {missing_env_vars}"

    def test_get_missing_vars(self) -> None:
        args = {
            **PROJECT_AND_CLUSTER,
            "LOGIN_FLOW": "client_credentials",
            "PROVIDER": "entra_id",
            "IDP_CLIENT_ID": "my-identifier",
        }

        with pytest.raises(ToolkitMissingValueError) as error:
            _ = EnvironmentVariables(**args).get_config(is_strict_validation=True)

        assert (
            str(error.value) == "The login flow 'client_credentials' requires the following environment variables: "
            "IDP_CLIENT_SECRET and IDP_TENANT_ID."
        )

    @mock.patch.dict(
        os.environ,
        {
            **PROJECT_AND_CLUSTER,
            "LOGIN_FLOW": "client_credentials",
            "IDP_CLIENT_ID": "my-identifier",
            "IDP_CLIENT_SECRET": "my-secret",
            "IDP_TENANT_ID": "my_tenant.onmicrosoft.com",
            "CDF_CLIENT_TIMEOUT": "10",
            "CDF_CLIENT_MAX_WORKERS": "5000",
        },
    )
    def test_create_environment_variables(
        self,
    ) -> None:
        env_vars = EnvironmentVariables.create_from_environment()

        assert env_vars.LOGIN_FLOW == "client_credentials"
        assert env_vars.CDF_CLUSTER == "toolkit-cluster"
        assert env_vars.CDF_PROJECT == "the-toolkit-project"
        assert env_vars.IDP_CLIENT_ID == "my-identifier"
        assert env_vars.IDP_CLIENT_SECRET == "my-secret"
        assert env_vars.IDP_TENANT_ID == "my_tenant.onmicrosoft.com"
        assert env_vars.CDF_CLIENT_TIMEOUT == 10
        assert env_vars.CDF_CLIENT_MAX_WORKERS == 5000

    def test_create_dot_env_file(self) -> None:
        env_vars = EnvironmentVariables(
            "bluefield",
            "cognite-toolkit",
            "entra_id",
            "client_credentials",
            IDP_CLIENT_ID="my-identifier",
            IDP_CLIENT_SECRET="my-secret",
            IDP_TENANT_ID="my_tenant.onmicrosoft.com",
        )

        assert (
            env_vars.create_dotenv_file()
            == """# .env file generated by cognite-toolkit
CDF_CLUSTER=bluefield
CDF_PROJECT=cognite-toolkit
PROVIDER=entra_id
LOGIN_FLOW=client_credentials

# Required variables
IDP_CLIENT_ID=my-identifier
IDP_CLIENT_SECRET=my-secret
IDP_TENANT_ID=my_tenant.onmicrosoft.com

# Optional variables (derived from the required variables)
CDF_URL=https://bluefield.cognitedata.com
IDP_TOKEN_URL=https://login.microsoftonline.com/my_tenant.onmicrosoft.com/oauth2/v2.0/token
IDP_AUDIENCE=https://bluefield.cognitedata.com
IDP_SCOPES=https://bluefield.cognitedata.com/.default
CDF_CLIENT_TIMEOUT=30
CDF_CLIENT_MAX_WORKERS=5
"""
        )



import difflib
import os
import sys
from pathlib import Path
from typing import Any
from unittest.mock import patch

import pytest
from cognite.client.data_classes.data_modeling import NodeId
from mypy.checkexpr import defaultdict
from rich import print
from rich.panel import Panel

from cognite_toolkit._cdf_tk.commands import BuildCommand, DeployCommand, PullCommand
from cognite_toolkit._cdf_tk.cruds import (
    RESOURCE_CRUD_LIST,
    FunctionCRUD,
    FunctionScheduleCRUD,
    GraphQLCRUD,
    HostedExtractorDestinationCRUD,
    HostedExtractorSourceCRUD,
    ResourceCRUD,
    ResourceWorker,
    SearchConfigCRUD,
    StreamlitCRUD,
    TransformationCRUD,
    WorkflowTriggerCRUD,
)
from cognite_toolkit._cdf_tk.cruds._resource_cruds.location import LocationFilterCRUD
from cognite_toolkit._cdf_tk.data_classes import BuiltModuleList, ResourceDeployResult
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from cognite_toolkit._cdf_tk.utils.file import remove_trailing_newline
from tests import data


@pytest.mark.skipif(
    sys.version_info < (3, 11), reason="We only run this test on Python 3.11+ to avoid parallelism issues"
)
def test_deploy_complete_org(env_vars: EnvironmentVariables, build_dir: Path) -> None:
    build = BuildCommand(silent=True, skip_tracking=True)

    built_modules = build.execute(
        verbose=False,
        organization_dir=data.COMPLETE_ORG,
        build_dir=build_dir,
        build_env_name="dev",
        no_clean=False,
        selected=None,
        client=env_vars.get_client(),
    )

    deploy_command = DeployCommand(silent=False, skip_tracking=True)
    client_id = os.environ["IDP_CLIENT_ID"]
    client_secret = os.environ["IDP_CLIENT_SECRET"]
    with patch.dict(
        os.environ,
        {"EVENTHUB_CLIENT_ID": client_id, "EVENTHUB_CLIENT_SECRET": client_secret},
    ):
        deploy_command.deploy_build_directory(
            env_vars=env_vars,
            build_dir=build_dir,
            build_env_name="dev",
            dry_run=False,
            drop=False,
            drop_data=False,
            force_update=False,
            include=None,
            verbose=True,
        )

    changed_resources = get_changed_resources(env_vars, build_dir)
    assert not changed_resources, "Redeploying the same resources should not change anything"

    changed_source_files = get_changed_source_files(env_vars, build_dir, built_modules, verbose=True)
    assert len(changed_source_files) == 0, f"Pulling the same source should not change anything {changed_source_files}"


@pytest.mark.skipif(
    sys.version_info < (3, 11), reason="We only run this test on Python 3.11+ to avoid parallelism issues"
)
def test_deploy_complete_org_alpha(env_vars: EnvironmentVariables, build_dir: Path) -> None:
    build = BuildCommand(silent=True, skip_tracking=True)

    built_modules = build.execute(
        verbose=False,
        organization_dir=data.COMPLETE_ORG_ALPHA_FLAGS,
        build_dir=build_dir,
        build_env_name="dev",
        no_clean=False,
        selected=None,
        client=env_vars.get_client(),
    )

    deploy_command = DeployCommand(silent=False, skip_tracking=True)
    client_id = os.environ["IDP_CLIENT_ID"]
    client_secret = os.environ["IDP_CLIENT_SECRET"]
    with patch.dict(
        os.environ,
        {"EVENTHUB_CLIENT_ID": client_id, "EVENTHUB_CLIENT_SECRET": client_secret},
    ):
        deploy_command.deploy_build_directory(
            env_vars,
            build_dir=build_dir,
            build_env_name="dev",
            dry_run=False,
            drop=False,
            drop_data=False,
            force_update=False,
            include=None,
            verbose=True,
        )

    changed_resources = get_changed_resources(env_vars, build_dir)
    assert not changed_resources, "Redeploying the same resources should not change anything"

    changed_source_files = get_changed_source_files(env_vars, build_dir, built_modules, verbose=False)
    assert not changed_source_files, "Pulling the same source should not change anything"


def get_changed_resources(env_vars: EnvironmentVariables, build_dir: Path) -> dict[str, set[Any]]:
    changed_resources: dict[str, set[Any]] = {}
    client = env_vars.get_client()
    for loader_cls in RESOURCE_CRUD_LIST:
        if loader_cls in {HostedExtractorSourceCRUD, HostedExtractorDestinationCRUD}:
            # These resources we have no way of knowing if they have changed. So they are always redeployed.
            continue
        loader = loader_cls.create_loader(client, build_dir)
        worker = ResourceWorker(loader, "deploy")
        files = worker.load_files()
        resources = worker.prepare_resources(files, environment_variables=env_vars.dump())
        if changed := (set(loader.get_ids(resources.to_update)) - {NodeId("sp_nodes", "MyExtendedFile")}):
            # We do not have a way to get CogniteFile extensions. This is a workaround to avoid the test failing.
            changed_resources[loader.display_name] = changed

    return changed_resources


def get_changed_source_files(
    env_vars: EnvironmentVariables, build_dir: Path, built_modules: BuiltModuleList, verbose: bool = False
) -> dict[str, set[Path]]:
    # This is a modified copy of the PullCommand._pull_build_dir and PullCommand._pull_resources methods
    # This will likely be hard to maintain, but if the pull command changes, should be refactored to be more
    # maintainable.
    cmd = PullCommand(silent=True, skip_tracking=True)
    changed_source_files: dict[str, set[str]] = defaultdict(set)
    selected_loaders = cmd._clean_command.get_selected_loaders(build_dir, read_resource_folders=set(), include=None)
    for loader_cls in selected_loaders:
        if (not issubclass(loader_cls, ResourceCRUD)) or (
            # Authentication that causes the diff to fail
            loader_cls in {HostedExtractorSourceCRUD, HostedExtractorDestinationCRUD}
            # External files that cannot (or not yet supported) be pulled
            or loader_cls in {GraphQLCRUD, FunctionCRUD, StreamlitCRUD}
            # Have authentication hashes that is different for each environment
            or loader_cls in {TransformationCRUD, FunctionScheduleCRUD, WorkflowTriggerCRUD}
            # LocationFilterLoader needs to split the file into multiple files, so we cannot compare them
            or loader_cls is LocationFilterCRUD
            # SearchConfigLoader is not supported in pull and post that also will require special handling
            or loader_cls is SearchConfigCRUD
        ):
            continue
        loader = loader_cls.create_loader(env_vars.get_client(), build_dir)
        resources = built_modules.get_resources(
            None, loader.folder_name, loader.kind, is_supported_file=loader.is_supported_file
        )
        if not resources:
            continue
        cdf_resources = loader.retrieve(resources.identifiers)
        cdf_resource_by_id = {loader.get_id(r): r for r in cdf_resources}

        resources_by_file = resources.by_file()
        file_results = ResourceDeployResult(loader.display_name)

        environment_variables = env_vars.dump()

        for source_file, resources in resources_by_file.items():
            if source_file.name == "extended.CogniteFile.yaml":
                # The extension of CogniteFile is not yet supported in Toolkit even though we have a test case for it.
                continue
            original_content = remove_trailing_newline(source_file.read_text())
            if "$FILENAME" in original_content:
                # File expansion pattern are not supported in pull.
                continue
            local_resource_by_id = cmd._get_local_resource_dict_by_id(resources, loader, environment_variables)
            _, to_write = cmd._get_to_write(local_resource_by_id, cdf_resource_by_id, file_results, loader)

            new_content, extra_files = cmd._to_write_content(
                source=original_content,
                to_write=to_write,
                resources=resources,
                environment_variables=environment_variables,
                loader=loader,
                source_file=source_file,
            )
            new_content = remove_trailing_newline(new_content)
            if new_content != original_content:
                if verbose:
                    print(
                        Panel(
                            "\n".join(difflib.unified_diff(original_content.splitlines(), new_content.splitlines())),
                            title=f"Diff for {source_file.name}",
                        )
                    )
                changed_source_files[loader.display_name].add(f"{loader.folder_name}/{source_file.name}")
            for path, new_extra_content in extra_files.items():
                new_extra_content = remove_trailing_newline(new_extra_content)
                original_extra_content = remove_trailing_newline(path.read_text(encoding="utf-8"))
                if new_extra_content != original_extra_content:
                    if verbose:
                        print(
                            Panel(
                                "\n".join(
                                    difflib.unified_diff(original_content.splitlines(), new_content.splitlines())
                                ),
                                title=f"Diff for {path.name}",
                            )
                        )
                    changed_source_files[loader.display_name].add(f"{loader.folder_name}/{path.name}")

    return dict(changed_source_files)



from collections.abc import Iterable, Sequence
from uuid import uuid4

from cognite.client.data_classes import Row, RowWrite

from cognite_toolkit._cdf_tk.cruds import RawDatabaseCRUD, RawTableCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.utils import sanitize_filename
from cognite_toolkit._cdf_tk.utils.http_client import HTTPClient, HTTPMessage, ItemsRequest
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal

from ._base import (
    ConfigurableStorageIO,
    Page,
    StorageIOConfig,
    TableUploadableStorageIO,
    UploadItem,
)
from .selectors import RawTableSelector


class RawIO(
    ConfigurableStorageIO[RawTableSelector, Row],
    TableUploadableStorageIO[RawTableSelector, Row, RowWrite],
):
    KIND = "RawRows"
    DISPLAY_NAME = "Raw Rows"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".yaml", ".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".parquet", ".csv", ".ndjson", ".yaml"})
    CHUNK_SIZE = 10_000
    UPLOAD_ENDPOINT = "/raw/dbs/{dbName}/tables/{tableName}/rows"
    BASE_SELECTOR = RawTableSelector

    def as_id(self, item: Row) -> str:
        return str(item.key)

    def count(self, selector: RawTableSelector) -> int | None:
        # Raw tables do not support aggregation queries, so we do not know the count
        # up front.
        return None

    def stream_data(self, selector: RawTableSelector, limit: int | None = None) -> Iterable[Page]:
        for chunk in self.client.raw.rows(
            db_name=selector.table.db_name,
            table_name=selector.table.table_name,
            limit=limit,
            # We cannot use partitions here as it is not thread safe. This spawn multiple threads
            # that are not shut down until all data is downloaded. We need to be able to abort.
            partitions=None,
            chunk_size=self.CHUNK_SIZE,
        ):
            yield Page(worker_id="main", items=chunk)

    def upload_items(
        self,
        data_chunk: Sequence[UploadItem[RowWrite]],
        http_client: HTTPClient,
        selector: RawTableSelector | None = None,
    ) -> Sequence[HTTPMessage]:
        if selector is None:
            raise ToolkitValueError("Selector must be provided for RawIO upload_items")
        url = self.UPLOAD_ENDPOINT.format(dbName=selector.table.db_name, tableName=selector.table.table_name)
        config = http_client.config
        return http_client.request_with_retries(
            message=ItemsRequest(
                endpoint_url=config.create_api_url(url),
                method="POST",
                items=list(data_chunk),
            )
        )

    def data_to_json_chunk(
        self, data_chunk: Sequence[Row], selector: RawTableSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        return [row.as_write().dump() for row in data_chunk]

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> RowWrite:
        return RowWrite._load(item_json)

    def configurations(self, selector: RawTableSelector) -> Iterable[StorageIOConfig]:
        yield StorageIOConfig(
            kind=RawDatabaseCRUD.kind,
            folder_name=RawDatabaseCRUD.folder_name,
            value={"dbName": selector.table.db_name},
            filename=sanitize_filename(selector.table.db_name),
        )
        yield StorageIOConfig(
            kind=RawTableCRUD.kind, folder_name=RawTableCRUD.folder_name, value=selector.table.model_dump(by_alias=True)
        )

    def row_to_resource(
        self, source_id: str, row: dict[str, JsonVal], selector: RawTableSelector | None = None
    ) -> RowWrite:
        key = str(uuid4())
        if selector is not None and selector.key is not None and selector.key in row:
            key = str(row.pop(selector.key))
        return RowWrite(key=key, columns=row)



from collections.abc import Hashable
from datetime import date, datetime
from typing import Any, Literal, TypeAlias, TypeVar, get_args

from cognite.client.data_classes import Annotation, Asset, Event, FileMetadata, TimeSeries
from cognite.client.data_classes._base import CogniteObject, WriteableCogniteResourceList

JsonVal: TypeAlias = None | str | int | float | bool | dict[str, "JsonVal"] | list["JsonVal"]

AssetCentricDestinationType: TypeAlias = Literal["assets", "files", "events", "timeseries", "sequences"]
AssetCentricType: TypeAlias = Literal["asset", "file", "event", "timeseries", "sequence", "fileAnnotation"]
AssetCentricResource: TypeAlias = Asset | FileMetadata | Event | TimeSeries
AssetCentricResourceExtended: TypeAlias = Asset | FileMetadata | Event | TimeSeries | Annotation
AssetCentricKind: TypeAlias = Literal["Assets", "Events", "TimeSeries", "FileMetadata", "FileAnnotations"]

DataType: TypeAlias = Literal["string", "integer", "float", "boolean", "json", "date", "timestamp", "epoch"]
PythonTypes: TypeAlias = str | int | float | bool | datetime | date | dict[str, Any] | list[Any]

AVAILABLE_DATA_TYPES: set[DataType] = set(get_args(DataType))

T_ID = TypeVar("T_ID", bound=Hashable)
T_WritableCogniteResourceList = TypeVar("T_WritableCogniteResourceList", bound=WriteableCogniteResourceList)
T_Value = TypeVar("T_Value")
PrimitiveType: TypeAlias = str | int | float | bool

T_WriteCogniteResource = TypeVar("T_WriteCogniteResource", bound=CogniteObject)
T_AssetCentricResource = TypeVar("T_AssetCentricResource", bound=AssetCentricResource)



import itertools
import sys
from datetime import datetime
from typing import Any, Literal

from cognite.client import CogniteClient
from cognite.client.data_classes._base import CogniteResourceList, WriteableCogniteResourceList
from cognite.client.data_classes.data_modeling import DirectRelationReference, ViewId
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteFile, CogniteFileApply
from cognite.client.utils._text import to_camel_case

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class ExtendableCogniteFileApply(CogniteFileApply):
    def __init__(
        self,
        space: str,
        external_id: str,
        *,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        source_id: str | None = None,
        source_context: str | None = None,
        source: DirectRelationReference | tuple[str, str] | None = None,
        source_created_time: datetime | None = None,
        source_updated_time: datetime | None = None,
        source_created_user: str | None = None,
        source_updated_user: str | None = None,
        assets: list[DirectRelationReference | tuple[str, str]] | None = None,
        mime_type: str | None = None,
        directory: str | None = None,
        category: DirectRelationReference | tuple[str, str] | None = None,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
        node_source: ViewId | None = None,
        extra_properties: dict[str, Any] | None = None,
    ) -> None:
        super().__init__(
            space=space,
            external_id=external_id,
            name=name,
            description=description,
            tags=tags,
            aliases=aliases,
            source_id=source_id,
            source_context=source_context,
            source=source,
            source_created_time=source_created_time,
            source_updated_time=source_updated_time,
            source_created_user=source_created_user,
            source_updated_user=source_updated_user,
            assets=assets,
            mime_type=mime_type,
            directory=directory,
            category=category,
            existing_version=existing_version,
            type=type,
        )
        self.node_source = node_source
        self.extra_properties = extra_properties

    def dump(self, camel_case: bool = True, context: Literal["api", "local"] = "api") -> dict[str, Any]:
        """Dumps the object to a dictionary.

        Args:
            camel_case: Whether to use camel case or not.
            context: If 'api', the output is for the API and will match the Node API schema. If 'local', the output is
                for a YAML file and all properties are  on the same level as the node properties. See below

        Example:
            >>> node = ExtendableCogniteFileApply(space="space", external_id="external_id", name="name")
            >>> node.dump(camel_case=True, context="api")
            {
                "space": "space",
                "externalId": "external_id",
                "sources": [
                    {
                        "source": {
                            "space": "cdf_cdm",
                            "externalId": "CogniteFile",
                            "version": "v1",
                            "type": "view"
                        }
                        "properties": {
                            "name": "name"
                        }
                    }
                ]
            }
            >>> node.dump(camel_case=True, context="local")
            {
                "space": "space",
                "external_id": "external_id",
                "name": "name",
            }

        Returns:

        """
        output = super().dump(camel_case)
        source = output["sources"][0]
        source["properties"].pop("node_source", None)
        source["properties"].pop("extra_properties", None)
        if context == "api":
            if self.node_source is not None:
                source["source"] = self.node_source.dump(include_type=True)
            if self.extra_properties is not None:
                source["properties"].update(self.extra_properties)
        else:
            output.pop("sources", None)
            output.update(source["properties"])
            if self.node_source is not None:
                output["nodeSource" if camel_case else "node_source"] = self.node_source.dump(include_type=False)
            if self.extra_properties is not None:
                output.update(self.extra_properties)
        return output

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        base_props = cls._load_base_properties(resource)
        properties = cls._load_properties(resource)
        loaded_keys = {to_camel_case(p) for p in itertools.chain(base_props.keys(), properties.keys())} | {
            "instanceType",
            "isUploaded",
            "uploadedTime",
        }
        if "nodeSource" in resource:
            properties["node_source"] = ViewId.load(resource["nodeSource"])
            loaded_keys.add("nodeSource")
        if extra_keys := (set(resource) - loaded_keys):
            properties["extra_properties"] = {key: resource[key] for key in extra_keys}

        return cls(**base_props, **properties)


class ExtendableCogniteFile(CogniteFile):
    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        *,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        source_id: str | None = None,
        source_context: str | None = None,
        source: DirectRelationReference | None = None,
        source_created_time: datetime | None = None,
        source_updated_time: datetime | None = None,
        source_created_user: str | None = None,
        source_updated_user: str | None = None,
        assets: list[DirectRelationReference] | None = None,
        mime_type: str | None = None,
        directory: str | None = None,
        is_uploaded: bool | None = None,
        uploaded_time: datetime | None = None,
        category: DirectRelationReference | None = None,
        type: DirectRelationReference | None = None,
        deleted_time: int | None = None,
        extra_properties: dict[str, Any] | None = None,
    ) -> None:
        super().__init__(
            space=space,
            external_id=external_id,
            version=version,
            last_updated_time=last_updated_time,
            created_time=created_time,
            name=name,
            description=description,
            tags=tags,
            aliases=aliases,
            source_id=source_id,
            source_context=source_context,
            source=source,
            source_created_time=source_created_time,
            source_updated_time=source_updated_time,
            source_created_user=source_created_user,
            source_updated_user=source_updated_user,
            assets=assets,
            mime_type=mime_type,
            directory=directory,
            is_uploaded=is_uploaded,
            uploaded_time=uploaded_time,
            category=category,
            type=type,
            deleted_time=deleted_time,
        )
        self.extra_properties = extra_properties

    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId(space="cdf_cdm", external_id="CogniteFile", version="v1")

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        base_props = cls._load_base_properties(resource)
        all_properties = resource.get("properties", {})
        # There should only be one source in one view
        if all_properties:
            view_props = next(iter(all_properties.values()))
            node_props = next(iter(view_props.values()))
            properties = cls._load_properties(node_props)
        else:
            properties = {}
        return cls(**base_props, **properties)

    def as_write(self) -> ExtendableCogniteFileApply:
        return ExtendableCogniteFileApply(
            space=self.space,
            external_id=self.external_id,
            name=self.name,
            description=self.description,
            tags=self.tags,
            aliases=self.aliases,
            source_id=self.source_id,
            source_context=self.source_context,
            source=self.source,
            source_created_time=self.source_created_time,
            source_updated_time=self.source_updated_time,
            source_created_user=self.source_created_user,
            source_updated_user=self.source_updated_user,
            assets=self.assets,  # type: ignore[arg-type]
            mime_type=self.mime_type,
            directory=self.directory,
            category=self.category,
            existing_version=self.version,
            type=self.type,
            node_source=None,
            extra_properties=self.extra_properties,
        )


class ExtendableCogniteFileApplyList(CogniteResourceList[ExtendableCogniteFileApply]):
    _RESOURCE = ExtendableCogniteFileApply


class ExtendableCogniteFileList(WriteableCogniteResourceList[ExtendableCogniteFileApply, ExtendableCogniteFile]):
    _RESOURCE = ExtendableCogniteFile

    def as_write(self) -> ExtendableCogniteFileApplyList:
        return ExtendableCogniteFileApplyList([model.as_write() for model in self.data])



# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
from collections.abc import Hashable, Iterable, Sequence
from typing import Any, final

from cognite.client.data_classes import (
    DataSet,
    DataSetList,
    DataSetWrite,
    DataSetWriteList,
    LabelDefinition,
    LabelDefinitionList,
    LabelDefinitionWrite,
    capabilities,
)
from cognite.client.data_classes.capabilities import (
    Capability,
    DataSetsAcl,
)
from cognite.client.data_classes.labels import LabelDefinitionWriteList
from cognite.client.exceptions import CogniteAPIError, CogniteDuplicatedError, CogniteNotFoundError
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitRequiredValueError,
)
from cognite_toolkit._cdf_tk.resource_classes import DataSetYAML, LabelsYAML

from .auth import GroupAllScopedCRUD


@final
class DataSetsCRUD(ResourceCRUD[str, DataSetWrite, DataSet, DataSetWriteList, DataSetList]):
    support_drop = False
    folder_name = "data_sets"
    resource_cls = DataSet
    resource_write_cls = DataSetWrite
    list_cls = DataSetList
    list_write_cls = DataSetWriteList
    yaml_cls = DataSetYAML
    kind = "DataSet"
    dependencies = frozenset({GroupAllScopedCRUD})
    _doc_url = "Data-sets/operation/createDataSets"

    @property
    def display_name(self) -> str:
        return "data sets"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[DataSetWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [DataSetsAcl.Action.Read]
            if read_only
            else [DataSetsAcl.Action.Read, DataSetsAcl.Action.Write, DataSetsAcl.Action.Owner]
        )

        return DataSetsAcl(
            actions,
            DataSetsAcl.Scope.All(),
        )

    @classmethod
    def get_id(cls, item: DataSet | DataSetWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if item.external_id is None:
            raise ToolkitRequiredValueError("DataSet must have external_id set.")
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> DataSetWrite:
        if resource.get("metadata"):
            for key, value in list(resource["metadata"].items()):
                if isinstance(value, dict | list):
                    resource["metadata"][key] = json.dumps(value)
        return DataSetWrite._load(resource)

    def dump_resource(self, resource: DataSet, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if "writeProtected" not in local and dumped.get("writeProtected") is False:
            # Default value is False, so we don't need to dump it.
            dumped.pop("writeProtected")
        if "metadata" not in local and not dumped.get("metadata"):
            # Default value is empty dict, so we don't need to dump it.
            dumped.pop("metadata", None)
        if "metadata" in dumped and "metadata" in local:
            meta_local = local["metadata"]
            for key, value in list(dumped["metadata"].items()):
                if isinstance(meta_local.get(key), dict | list):
                    try:
                        converted = json.loads(value)
                    except json.JSONDecodeError:
                        continue
                    dumped["metadata"][key] = converted

        return dumped

    def create(self, items: Sequence[DataSetWrite]) -> DataSetList:
        items = list(items)
        created = DataSetList([], cognite_client=self.client)
        # There is a bug in the data set API, so only one duplicated data set is returned at the time,
        # so we need to iterate.
        while len(items) > 0:
            try:
                created.extend(DataSetList(self.client.data_sets.create(items)))
                return created
            except CogniteDuplicatedError as e:
                if len(e.duplicated) < len(items):
                    for dup in e.duplicated:
                        ext_id = dup.get("externalId", None)
                        for item in items:
                            if item.external_id == ext_id:
                                items.remove(item)
                else:
                    items = []
        return created

    def retrieve(self, ids: SequenceNotStr[str]) -> DataSetList:
        return self.client.data_sets.retrieve_multiple(external_ids=ids, ignore_unknown_ids=True)

    def update(self, items: DataSetWriteList) -> DataSetList:
        return self.client.data_sets.update(items, mode="replace")

    def delete(self, ids: SequenceNotStr[str]) -> int:
        raise NotImplementedError("CDF does not support deleting data sets.")

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[DataSet]:
        return iter(self.client.data_sets)


@final
class LabelCRUD(
    ResourceCRUD[str, LabelDefinitionWrite, LabelDefinition, LabelDefinitionWriteList, LabelDefinitionList]
):
    folder_name = "classic"
    filename_pattern = r"^.*Label$"  # Matches all yaml files whose stem ends with *Label.
    resource_cls = LabelDefinition
    resource_write_cls = LabelDefinitionWrite
    list_cls = LabelDefinitionList
    list_write_cls = LabelDefinitionWriteList
    yaml_cls = LabelsYAML
    kind = "Label"
    dependencies = frozenset({DataSetsCRUD, GroupAllScopedCRUD})
    _doc_url = "Labels/operation/createLabelDefinitions"
    support_update = False

    @property
    def display_name(self) -> str:
        return "labels"

    @classmethod
    def get_id(cls, item: LabelDefinition | LabelDefinitionWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if not item.external_id:
            raise ToolkitRequiredValueError("LabelDefinition must have external_id set.")
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[LabelDefinitionWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        scope: capabilities.LabelsAcl.Scope.All | capabilities.LabelsAcl.Scope.DataSet = (  # type: ignore[valid-type]
            capabilities.LabelsAcl.Scope.All()
        )
        if items:
            if data_set_ids := {item.data_set_id for item in items if item.data_set_id}:
                scope = capabilities.LabelsAcl.Scope.DataSet(list(data_set_ids))

        actions = (
            [capabilities.LabelsAcl.Action.Read]
            if read_only
            else [capabilities.LabelsAcl.Action.Read, capabilities.LabelsAcl.Action.Write]
        )

        return capabilities.LabelsAcl(actions, scope)

    def create(self, items: LabelDefinitionWriteList) -> LabelDefinitionList:
        return self.client.labels.create(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> LabelDefinitionList:
        return self.client.labels.retrieve(ids, ignore_unknown_ids=True)

    def delete(self, ids: SequenceNotStr[str]) -> int:
        try:
            self.client.labels.delete(ids)
        except (CogniteAPIError, CogniteNotFoundError) as e:
            non_existing = set(e.failed or [])
            if existing := [id_ for id_ in ids if id_ not in non_existing]:
                self.client.labels.delete(existing)
            return len(existing)
        else:
            # All deleted successfully
            return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[LabelDefinition]:
        return iter(self.client.labels(data_set_external_ids=[data_set_external_id] if data_set_external_id else None))

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        """Returns all items that this item requires.

        For example, a TimeSeries requires a DataSet, so this method would return the
        DatasetLoader and identifier of that dataset.
        """
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, item["dataSetExternalId"]

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> LabelDefinitionWrite:
        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        return LabelDefinitionWrite._load(resource)

    def dump_resource(self, resource: LabelDefinition, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        return dumped
